"""Module for commonly used validator methods for protein references."""
from typing import List

from uta_tools.data_sources import SeqRepoAccess

from variation.tokenizers.caches import AminoAcidCache


class ProteinBase:
    """Protein Base class."""

    def __init__(self, seq_repo_access: SeqRepoAccess,
                 amino_acid_cache: AminoAcidCache) -> None:
        """Initialize Protein Base class.

        :param SeqRepoAccess seq_repo_access: Access to SeqRepo data
        :param AminoAcidCache amino_acid_cache: Amino acid code data
        """
        self.seqrepo_access = seq_repo_access
        self.amino_acid_cache = amino_acid_cache

    def check_ref_aa(self, t: str, aa: str, pos: int, errors: List) -> None:
        """Check that reference protein matches actual protein.

        :param str t: Transcript
        :param str aa: Expected Protein
        :param int pos: Expected position
        :param List errors: List of errors
        """
        ref_aa, warning = self.seqrepo_access.get_reference_sequence(t, pos)
        if not ref_aa:
            errors.append(warning)
        else:
            if ref_aa and len(ref_aa) == 1 and len(aa) == 3:
                aa = self.amino_acid_cache.convert_three_to_one(aa)

            if ref_aa != aa:
                errors.append(f"Needed to find {aa} at position {pos} on {t} but found"
                              f" {ref_aa}")



"""Module for Classification schema."""
from typing import List
from enum import IntEnum

from pydantic import BaseModel

from variation.schemas.token_response_schema import Token


class ClassificationType(IntEnum):
    """Enums for Classification Types."""

    FUSION = 1
    PROTEIN_SUBSTITUTION = 2
    POLYPEPTIDE_TRUNCATION = 3
    SILENT_MUTATION = 4
    PROTEIN_FRAMESHIFT = 5
    PROTEIN_ALTERNATE = 6
    PROTEIN_DELINS = 7
    PROTEIN_TERMINATION = 8
    PROTEIN_DUPLICATION = 9
    ONCOGENIC = 10
    EXPRESSION = 11
    COMPLEX = 12
    CODING_DNA_SUBSTITUTION = 13
    GENOMIC_SUBSTITUTION = 14
    CODING_DNA_SILENT_MUTATION = 15
    GENOMIC_SILENT_MUTATION = 16
    CODING_DNA_DELINS = 17
    GENOMIC_DELINS = 18
    PROTEIN_DELETION = 19
    CODING_DNA_DELETION = 20
    GENOMIC_DELETION = 21
    PROTEIN_INSERTION = 22
    CODING_DNA_INSERTION = 23
    GENOMIC_INSERTION = 24
    GENOMIC_UNCERTAIN_DELETION = 25
    GENOMIC_DUPLICATION = 26
    GENOMIC_DELETION_RANGE = 27


class ConfidenceRating(IntEnum):
    """Enums for classification confidence ratings."""

    INTERSECTION = 1
    SUPERSET = 2
    UNORDERED = 3
    EXACT = 4


class Classification(BaseModel):
    """Classification for a list of tokens."""

    classification_type: ClassificationType
    matching_tokens: List[str]
    non_matching_tokens: List[str]
    all_tokens: List[Token]
    confidence: ConfidenceRating


class ClassificationResponseSchema(BaseModel):
    """Classification response for a given query."""

    search_term: str
    classifications: List[Classification]



"""Module for testing Genomic Deletion Classifier."""
import unittest

from variation.classifiers import GenomicDeletionClassifier
from .classifier_base import ClassifierBase


class TestGenomicDeletionClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Genomic Deletion Classifier."""

    def classifier_instance(self):
        """Return GenomicDeletionClassifier instance."""
        return GenomicDeletionClassifier()

    def fixture_name(self):
        """Return GenomicDeletionClassifier fixture name."""
        return "genomic_deletion"



"""A module for Genomic Insertion Tokenization."""
from typing import Dict, Optional

from variation.schemas.token_response_schema import GenomicInsertionToken
from variation.tokenizers.insertion_base import InsertionBase


class GenomicInsertion(InsertionBase):
    """Class for tokenizing Insertion at the genomic reference sequence."""

    def return_token(self, params: Dict) -> Optional[GenomicInsertionToken]:
        """Return Genomic Insertion token."""
        if self.parts["coordinate_type"] == "g":
            return GenomicInsertionToken(**params)



"""Module for testing Coding DNA Silent Mutation Translator."""
import unittest

from variation.classifiers import CodingDNASilentMutationClassifier
from variation.translators import CodingDNASilentMutation
from variation.validators import CodingDNASilentMutation as CDNASM_V
from .translator_base import TranslatorBase


class TestCodingDNASilentMutationTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the Coding DNA Silent Mutation Translator."""

    def classifier_instance(self):
        """Return coding DNA silent mutation instance."""
        return CodingDNASilentMutationClassifier()

    def validator_instance(self):
        """Return coding DNA silent mutation instance."""
        return CDNASM_V(*self.params)

    def translator_instance(self):
        """Return coding DNA silent mutation instance."""
        return CodingDNASilentMutation()

    def fixture_name(self):
        """Return the fixture name for coding DNA silent mutation."""
        return "coding_dna_silent_mutation"



"""Module for testing Coding DNA Insertion Validator."""
import unittest

from variation.validators import CodingDNAInsertion
from variation.classifiers import CodingDNAInsertionClassifier
from .validator_base import ValidatorBase


class TestCodingDNAInsertionValidator(ValidatorBase, unittest.TestCase):
    """A class to test the CodingDNAInsertion Validator."""

    def validator_instance(self):
        """Return coding dna insertion instance."""
        return CodingDNAInsertion(*self.params)

    def classifier_instance(self):
        """Return the coding dna insertion classifier instance."""
        return CodingDNAInsertionClassifier()

    def fixture_name(self):
        """Return the fixture name for coding dna insertion."""
        return "coding_dna_insertion"



"""Module for testing Protein Insertion Validator."""
import unittest

from variation.validators import ProteinInsertion
from variation.classifiers import ProteinInsertionClassifier
from .validator_base import ValidatorBase


class TestProteinInsertionValidator(ValidatorBase, unittest.TestCase):
    """A class to test the Protein Insertion Validator."""

    def validator_instance(self):
        """Return protein insertion instance."""
        return ProteinInsertion(*self.aa_params)

    def classifier_instance(self):
        """Return the protein insertion classifier instance."""
        return ProteinInsertionClassifier()

    def fixture_name(self):
        """Return the fixture name for protein insertion."""
        return "protein_insertion"



"""A module for the Protein Insertion Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class ProteinInsertionClassifier(SetBasedClassifier):
    """The Protein Insertion Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Protein Insertion classification type."""
        return ClassificationType.PROTEIN_INSERTION

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["ProteinInsertion"],
            ["GeneSymbol", "ProteinSubstitution", "ProteinInsertion"],
            ["ProteinInsertion", "GeneSymbol"],
            ["GeneSymbol", "ProteinInsertion"],
            ["HGVS", "ProteinInsertion"],
            ["ReferenceSequence", "ProteinInsertion"],
            ["LocusReferenceGenomic", "ProteinInsertion"]
        ]



"""The module for DelIns Validation."""
from typing import List, Dict, Optional
import logging

from ga4gh.vrsatile.pydantic.vrs_models import RelativeCopyClass

from variation.schemas.classification_response_schema import Classification, \
    ClassificationType
from variation.schemas.app_schemas import Endpoint
from variation.schemas.token_response_schema import Token, GeneMatchToken
from variation.validators.validator import Validator
from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum

logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)


class DelInsBase(Validator):
    """The DelIns Validator Base class."""

    def is_token_instance(self, t: Token) -> bool:
        """Check to see if token is instance of a token type.

        :param Token t: Classification token to find type of
        :return: `True` if token is instance of class token. `False` otherwise.
        """
        raise NotImplementedError

    def variation_name(self) -> str:
        """Return the variation name.

        :return: variation class name
        """
        raise NotImplementedError

    def get_gene_tokens(
            self, classification: Classification) -> List[GeneMatchToken]:
        """Return a list of gene tokens for a classification.

        :param Classification classification: Classification for a list of
            tokens
        :return: A list of gene tokens for the classification
        """
        raise NotImplementedError

    async def get_transcripts(self, gene_tokens: List, classification: Classification,
                              errors: List) -> Optional[List[str]]:
        """Get transcript accessions for a given classification.

        :param List gene_tokens: A list of gene tokens
        :param Classification classification: A classification for a list of
            tokens
        :param List errors: List of errors
        :return: List of transcript accessions
        """
        raise NotImplementedError

    def validates_classification_type(
            self, classification_type: ClassificationType) -> bool:
        """Check that classification type can be validated by validator.

        :param ClassificationType classification_type: The type of variation
        :return: `True` if classification_type matches validator's
            classification type. `False` otherwise.
        """
        raise NotImplementedError

    async def get_valid_invalid_results(
        self, classification_tokens: List, transcripts: List,
        classification: Classification, results: List, gene_tokens: List,
        mane_data_found: Dict, is_identifier: bool,
        hgvs_dup_del_mode: HGVSDupDelModeEnum,
        endpoint_name: Optional[Endpoint] = None,
        baseline_copies: Optional[int] = None,
        relative_copy_class: Optional[RelativeCopyClass] = None,
        do_liftover: bool = False
    ) -> None:
        """Add validation result objects to a list of results.

        :param List classification_tokens: A list of classification Tokens
        :param List transcripts: A list of transcript accessions
        :param Classification classification: A classification for a list of
            tokens
        :param List results: Stores validation result objects
        :param List gene_tokens: List of GeneMatchTokens for a classification
        :param Dict mane_data_found: MANE Transcript information found
        :param bool is_identifier: `True` if identifier is given for exact
            location. `False` otherwise.
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Optional[Endpoint] endpoint_name: Then name of the endpoint being used
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        """
        raise NotImplementedError

    def check_pos_index(self, t: str, s: Token, errors: List) -> None:
        """Check that position exists on transcript.

        :param str t: Transcript accession
        :param Token s: Classification token
        :param List errors: List of errors
        """
        if s.end_pos_del:
            sequence, w = self.seqrepo_access.get_reference_sequence(
                t, int(s.start_pos_del), int(s.end_pos_del))
        else:
            sequence, w = self.seqrepo_access.get_reference_sequence(
                t, int(s.start_pos_del))
        if sequence is None:
            errors.append(w)



"""Module for testing Coding DNA Substitution Classifier."""
import unittest

from variation.classifiers import CodingDNASubstitutionClassifier
from .classifier_base import ClassifierBase


class TestCodingDNASubstitutionClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Coding DNA Substitution Classifier."""

    def classifier_instance(self):
        """Return CodingDNASubstitutionClassifier instance."""
        return CodingDNASubstitutionClassifier()

    def fixture_name(self):
        """Return CodingDNASubstitutionClassifier fixture name."""
        return "coding_dna_substitution"



"""Module for testing Silent Mutation Validator."""
import unittest

from variation.validators import SilentMutation
from variation.classifiers import SilentMutationClassifier
from .validator_base import ValidatorBase


class TestSilentMutationValidator(ValidatorBase, unittest.TestCase):
    """A class to test the Silent Mutation Validator."""

    def validator_instance(self):
        """Return Silent Mutation instance."""
        return SilentMutation(*self.aa_params)

    def classifier_instance(self):
        """Return the Silent Mutation classifier instance."""
        return SilentMutationClassifier()

    def fixture_name(self):
        """Return the fixture name for Silent Mutation."""
        return "silent_mutation"



"""A module for the Genomic Uncertain Deletion Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class GenomicUncertainDeletionClassifier(SetBasedClassifier):
    """The Genomic Uncertain Deletion Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Genomic Uncertain Deletion classification type."""
        return ClassificationType.GENOMIC_UNCERTAIN_DELETION

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["Chromosome", "GenomicUncertainDeletion"],
            ["GeneSymbol", "GenomicUncertainDeletion"],
            ["GenomicUncertainDeletion", "GeneSymbol"],
            ["HGVS", "GenomicUncertainDeletion"],
            ["ReferenceSequence", "GenomicUncertainDeletion"],
            ["LocusReferenceGenomic", "GenomicUncertainDeletion"]
        ]



"""Module for testing Genomic Silent Mutation Translator."""
import unittest

from variation.classifiers import GenomicSilentMutationClassifier
from variation.translators import GenomicSilentMutation
from variation.validators import GenomicSilentMutation as GENOMICSM_V
from .translator_base import TranslatorBase


class TestGenomicSilentMutationTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the Genomic Silent Mutation Translator."""

    def classifier_instance(self):
        """Return genomic silent mutation instance."""
        return GenomicSilentMutationClassifier()

    def validator_instance(self):
        """Return genomic silent mutation instance."""
        return GENOMICSM_V(*self.params)

    def translator_instance(self):
        """Return genomic silent mutation instance."""
        return GenomicSilentMutation()

    def fixture_name(self):
        """Return the fixture name for genomic silent mutation."""
        return "genomic_silent_mutation"



"""Module for Codon Table."""
from typing import Dict, List

from variation.tokenizers.caches import AminoAcidCache


MULTIPLE_CODONS = {"Leu", "Ser", "Arg"}


class CodonTable:
    """Class for codon table data."""

    def __init__(self, amino_acid_cache: AminoAcidCache) -> None:
        """Initialize codon table class.

        :param AminoAcidCache amino_acid_cache: Amino acid code data
        """
        self.table = self._set_codon_table()
        self.amino_acid_cache = amino_acid_cache
        self._dna_to_rna = {
            "T": "A", "A": "U",
            "G": "C", "C": "G"
        }

    @staticmethod
    def _set_codon_table() -> Dict[str, str]:
        """Return RNA codon mappings for amino acids.

        :return: RNA codon --> amino acid single code mappings
        """
        return {
            "AUA": "I", "AUC": "I", "AUU": "I", "AUG": "M",
            "ACA": "T", "ACC": "T", "ACG": "T", "ACU": "T",
            "AAC": "N", "AAU": "N", "AAA": "K", "AAG": "K",
            "AGC": "S", "AGU": "S", "AGA": "R", "AGG": "R",
            "CUA": "L", "CUC": "L", "CUG": "L", "CUU": "L",
            "CCA": "P", "CCC": "P", "CCG": "P", "CCU": "P",
            "CAC": "H", "CAU": "H", "CAA": "Q", "CAG": "Q",
            "CGA": "R", "CGC": "R", "CGG": "R", "CGU": "R",
            "GUA": "V", "GUC": "V", "GUG": "V", "GUU": "V",
            "GCA": "A", "GCC": "A", "GCG": "A", "GCU": "A",
            "GAC": "D", "GAU": "D", "GAA": "E", "GAG": "E",
            "GGA": "G", "GGC": "G", "GGG": "G", "GGU": "G",
            "UCA": "S", "UCC": "S", "UCG": "S", "UCU": "S",
            "UUC": "F", "UUU": "F", "UUA": "L", "UUG": "L",
            "UAC": "Y", "UAU": "Y", "UAA": "*", "UAG": "*",
            "UGC": "C", "UGU": "C", "UGA": "*", "UGG": "W",
        }

    def get_codons(self, amino_acid: str) -> List[str]:
        """Return a list of codons for an amino acid.

        :param str amino_acid: Amino acid to get codons for
        :return: List of codons
        """
        amino_acid = amino_acid.upper()
        if len(amino_acid) == 3:
            amino_acid = self.amino_acid_cache.convert_three_to_one(amino_acid)

        codons = list()
        for k, v in self.table.items():
            if v == amino_acid:
                codons.append(k)
        codons.sort()
        return codons

    def dna_to_rna(self, dna_codon: str) -> str:
        """Convert DNA codon to RNA codon.

        :param str dna_codon: DNA codon
        :return: RNA codon
        """
        dna_codon_list = list(dna_codon)
        rna_codon = ""
        for char in dna_codon_list:
            rna_codon += self._dna_to_rna[char]
        return rna_codon



"""Module for testing Protein Deletion Classifier."""
import unittest

from variation.classifiers import ProteinDeletionClassifier
from .classifier_base import ClassifierBase


class TestProteinDeletionClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Protein Deletion Classifier."""

    def classifier_instance(self):
        """Return ProteinDeletionClassifier instance."""
        return ProteinDeletionClassifier()

    def fixture_name(self):
        """Return ProteinDeletionClassifier fixture name."""
        return "protein_deletion"



"""A module for testing gnomad VCF tokenization."""
import pytest

from variation.schemas.token_response_schema import ChromosomeToken, \
    GenomicDeletionToken, CoordinateType, GenomicSubstitutionToken, \
    GenomicInsertionToken, GenomicSilentMutationToken
from variation.tokenizers import GnomadVCF


@pytest.fixture(scope="module")
def tokenizer():
    """Create GnomadVCF Tokenizer test fixture."""
    return GnomadVCF()


def chr_assertion_checks(token, input_str, token_val, chromosome_val):
    """Chromosome token assertion checks."""
    assert isinstance(token, ChromosomeToken)
    assert token.token == token_val
    assert token.input_string == input_str
    assert token.chromosome == chromosome_val
    assert token.nomenclature == "gnomad_vcf"


def test_matches(tokenizer):
    """Test that valid input returns correct output"""
    # Genomic Substitution
    input_str = "7-140453136-A-T"
    tokens = tokenizer.match(input_str)
    assert tokens
    assert len(tokens) == 2
    chr_assertion_checks(tokens[0], input_str, "7", "chr7")
    token = tokens[1]
    assert isinstance(token, GenomicSubstitutionToken)
    assert token.position == 140453136
    assert token.ref_nucleotide == "A"
    assert token.new_nucleotide == "T"
    assert token.coordinate_type == CoordinateType.LINEAR_GENOMIC
    assert token.token_type == "GenomicSubstitution"
    assert token.alt_type == "substitution"
    assert token.so_id == "SO:0001483"
    assert token.molecule_context == "genomic"
    assert token.nomenclature == "gnomad_vcf"

    # Genomic Insertion
    input_str = "chromosome22-6548-G-GCTAG"
    tokens = tokenizer.match(input_str)
    assert tokens
    assert len(tokens) == 2
    chr_assertion_checks(tokens[0], input_str, "chromosome22", "chr22")
    token = tokens[1]
    assert isinstance(token, GenomicInsertionToken)
    assert token.start_pos_flank == 6548
    assert token.end_pos_flank == 6549
    assert token.inserted_sequence == "CTAG"
    assert token.inserted_sequence2 is None
    assert token.coordinate_type == CoordinateType.LINEAR_GENOMIC
    assert token.token_type == "GenomicInsertion"
    assert token.alt_type == "insertion"
    assert token.so_id == "SO:0000667"
    assert token.molecule_context == "genomic"
    assert token.nomenclature == "gnomad_vcf"

    # Genomic Deletion
    input_str = "x-1234-ac-a"
    tokens = tokenizer.match(input_str)
    assert tokens
    assert len(tokens) == 2
    chr_assertion_checks(tokens[0], input_str, "x", "chrX")
    token = tokens[1]
    assert isinstance(token, GenomicDeletionToken)
    assert token.start_pos_del == 1235
    assert token.end_pos_del == 1235
    assert token.coordinate_type == CoordinateType.LINEAR_GENOMIC
    assert token.deleted_sequence == "C"
    assert token.token_type == "GenomicDeletion"
    assert token.alt_type == "deletion"
    assert token.so_id == "SO:0000159"
    assert token.molecule_context == "genomic"
    assert token.nomenclature == "gnomad_vcf"

    # Genomic Silent Mutation
    input_str = "chrY-451-C-C"
    tokens = tokenizer.match(input_str)
    assert tokens
    assert len(tokens) == 2
    chr_assertion_checks(tokens[0], input_str, "chrY", "chrY")
    token = tokens[1]
    assert isinstance(token, GenomicSilentMutationToken)
    assert token.position == 451
    assert token.ref_nucleotide == "C"
    assert token.new_nucleotide == "="
    assert token.coordinate_type == CoordinateType.LINEAR_GENOMIC
    assert token.token_type == "GenomicSilentMutation"
    assert token.alt_type == "silent_mutation"
    assert token.so_id == "SO:0002073"
    assert token.molecule_context == "genomic"
    assert token.nomenclature == "gnomad_vcf"


def test_no_matches(tokenizer):
    """Test that invalid input returns no tokens"""
    assert tokenizer.match("") is None
    assert tokenizer.match("7-A-C-T") is None
    assert tokenizer.match("7-131-G-B") is None
    assert tokenizer.match("7-131-S-A") is None
    assert tokenizer.match("23-24-A-C") is None



"""Module for testing Genomic Insertion Validator."""
import unittest

from variation.validators import GenomicInsertion
from variation.classifiers import GenomicInsertionClassifier
from .validator_base import ValidatorBase


class TestGenomicInsertionValidator(ValidatorBase, unittest.TestCase):
    """A class to test the GenomicInsertion Validator."""

    def validator_instance(self):
        """Return genomic insertion instance."""
        return GenomicInsertion(*self.params)

    def classifier_instance(self):
        """Return the genomic insertion classifier instance."""
        return GenomicInsertionClassifier()

    def fixture_name(self):
        """Return the fixture name for genomic insertion."""
        return "genomic_insertion"



"""Create methods used throughout tests."""
import asyncio

import pytest
from ga4gh.vrsatile.pydantic.vrsatile_models import VariationDescriptor

from variation.query import QueryHandler


@pytest.fixture(scope="session")
def event_loop(request):
    """Create an instance of the default event loop for each test case."""
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
def test_query_handler():
    """Build normalize test fixture."""
    return QueryHandler()


@pytest.fixture(scope="session")
def vhl_gene_context():
    """Create a VHL gene context."""
    return {
        "id": "normalize.gene:VHL",
        "type": "GeneDescriptor",
        "label": "VHL",
        "gene_id": "hgnc:12687",
        "xrefs": [
            "ncbigene:7428",
            "ensembl:ENSG00000134086"
        ],
        "alternate_labels": [
            "HRCA1",
            "VHL1",
            "RCA1",
            "pVHL"
        ],
        "extensions": [
            {
                "type": "Extension",
                "name": "symbol_status",
                "value": "approved"
            },
            {
                "type": "Extension",
                "name": "approved_name",
                "value": "von Hippel-Lindau tumor suppressor"
            },
            {
                "type": "Extension",
                "name": "associated_with",
                "value": [
                    "ucsc:uc003bvc.4",
                    "pubmed:9671762",
                    "refseq:NM_000551",
                    "cosmic:VHL",
                    "omim:608537",
                    "vega:OTTHUMG00000128668",
                    "ccds:CCDS2598",
                    "ena.embl:L15409",
                    "orphanet:120467",
                    "ccds:CCDS2597",
                    "uniprot:P40337"
                ]
            },
            {
                "type": "Extension",
                "name": "chromosome_location",
                "value": {
                    "_id":
                        "ga4gh:VCL.S-TtMfLdsgZPVRrWEf1-jiZMyTDCt5y1",
                    "type": "ChromosomeLocation",
                    "species_id": "taxonomy:9606",
                    "chr": "3",
                    "interval": {
                        "end": "p25.3",
                        "start": "p25.3",
                        "type": "CytobandInterval"
                    }
                }
            },
            {
                "name": "previous_symbols",
                "value": [
                    "RCA1"
                ],
                "type": "Extension"
            }
        ]
    }


@pytest.fixture(scope="session")
def braf_gene_context():
    """Create BRAF gene context test fixture."""
    return {
        "id": "normalize.gene:BRAF",
        "type": "GeneDescriptor",
        "label": "BRAF",
        "gene_id": "hgnc:1097",
        "xrefs": [
            "ncbigene:673",
            "ensembl:ENSG00000157764"
        ],
        "alternate_labels": [
            "BRAF1",
            "RAFB1",
            "B-raf",
            "B-RAF1",
            "NS7"
        ],
        "extensions": [
            {
                "type": "Extension",
                "name": "symbol_status",
                "value": "approved"
            },
            {
                "type": "Extension",
                "name": "approved_name",
                "value": "B-Raf proto-oncogene, serine/threonine kinase"
            },
            {
                "type": "Extension",
                "name": "associated_with",
                "value": [
                    "ucsc:uc003vwc.5",
                    "pubmed:1565476",
                    "omim:164757",
                    "vega:OTTHUMG00000157457",
                    "ccds:CCDS5863",
                    "iuphar:1943",
                    "ccds:CCDS87555",
                    "orphanet:119066",
                    "refseq:NM_004333",
                    "ena.embl:M95712",
                    "pubmed:2284096",
                    "uniprot:P15056",
                    "cosmic:BRAF"
                ]
            },
            {
                "type": "Extension",
                "name": "chromosome_location",
                "value": {
                    "_id": "ga4gh:VCL.O6yCQ1cnThOrTfK9YUgMlTfM6HTqbrKw",
                    "type": "ChromosomeLocation",
                    "species_id": "taxonomy:9606",
                    "chr": "7",
                    "interval": {
                        "end": "q34",
                        "start": "q34",
                        "type": "CytobandInterval"
                    }
                }
            }
        ]
    }


@pytest.fixture(scope="session")
def egfr_context():
    """Create EGFR gene context test fixture"""
    return {
        "id": "normalize.gene:EGFR",
        "type": "GeneDescriptor",
        "label": "EGFR",
        "gene_id": "hgnc:3236",
        "xrefs": [
            "ncbigene:1956",
            "ensembl:ENSG00000146648"
        ],
        "alternate_labels": [
            "HER1",
            "NISBD2",
            "ERBB",
            "PIG61",
            "mENA",
            "ERBB1",
            "ERRP"
        ],
        "extensions": [
            {
                "type": "Extension",
                "name": "symbol_status",
                "value": "approved"
            },
            {
                "name": "approved_name",
                "value": "epidermal growth factor receptor",
                "type": "Extension"
            },
            {
                "type": "Extension",
                "name": "associated_with",
                "value": [
                    "ccds:CCDS5516",
                    "ccds:CCDS5514",
                    "iuphar:1797",
                    "uniprot:P00533",
                    "vega:OTTHUMG00000023661",
                    "ucsc:uc003tqk.4",
                    "ccds:CCDS5515",
                    "refseq:NM_005228",
                    "ccds:CCDS87506",
                    "ccds:CCDS47587",
                    "pubmed:1505215",
                    "cosmic:EGFR",
                    "ccds:CCDS87507",
                    "omim:131550",
                    "orphanet:121311"
                ]
            },
            {
                "type": "Extension",
                "name": "chromosome_location",
                "value": {
                    "_id": "ga4gh:VCL.wgFi9e72ZIIJaOfLx5gaOeGrwP_IZoQ2",
                    "type": "ChromosomeLocation",
                    "species_id": "taxonomy:9606",
                    "chr": "7",
                    "interval": {
                        "end": "p11.2",
                        "start": "p11.2",
                        "type": "CytobandInterval"
                    }
                }
            },
            {
                "name": "previous_symbols",
                "value": [
                    "ERBB"
                ],
                "type": "Extension"
            }
        ]
    }


@pytest.fixture(scope="session")
def erbb2_context():
    """Create test fixture for ERBB2 Gene Context."""
    return {
        "id": "normalize.gene:ERBB2",
        "type": "GeneDescriptor",
        "label": "ERBB2",
        "gene_id": "hgnc:3430",
        "xrefs": [
            "ncbigene:2064",
            "ensembl:ENSG00000141736"
        ],
        "alternate_labels": [
            "NGL",
            "CD340",
            "HER2",
            "NEU",
            "TKR1",
            "HER-2",
            "HER-2/neu",
            "VSCN2",
            "MLN 19"
        ],
        "extensions": [
            {
                "type": "Extension",
                "name": "symbol_status",
                "value": "approved"
            },
            {
                "name": "approved_name",
                "value": "erb-b2 receptor tyrosine kinase 2",
                "type": "Extension"
            },
            {
                "type": "Extension",
                "name": "associated_with",
                "value": [
                    "ucsc:uc002hso.4",
                    "ena.embl:X03363",
                    "ccds:CCDS77017",
                    "vega:OTTHUMG00000179300",
                    "ccds:CCDS77016",
                    "uniprot:P04626",
                    "refseq:NM_004448",
                    "ccds:CCDS74052",
                    "hcdmdb:CD340",
                    "omim:164870",
                    "ccds:CCDS32642",
                    "ccds:CCDS45667",
                    "cosmic:ERBB2",
                    "iuphar:2019"
                ]
            },
            {
                "type": "Extension",
                "name": "chromosome_location",
                "value": {
                    "_id": "ga4gh:VCL.pS7M3aeNymozN9LKeAwVDEB5H1nt4Kqy",
                    "type": "ChromosomeLocation",
                    "species_id": "taxonomy:9606",
                    "chr": "17",
                    "interval": {
                        "end": "q12",
                        "start": "q12",
                        "type": "CytobandInterval"
                    }
                }
            },
            {
                "name": "previous_symbols",
                "value": [
                    "NGL"
                ],
                "type": "Extension"
            }
        ]
    }


@pytest.fixture(scope="session")
def braf_600loc():
    """Create test fixture for BRAF 600 location"""
    return {
        "_id": "ga4gh:VSL.2cHIgn7iLKk4x9z3zLkSTTFMV0e48DR4",
        "interval": {
            "end": {"value": 600, "type": "Number"},
            "start": {"value": 599, "type": "Number"},
            "type": "SequenceInterval"
        },
        "sequence_id": "ga4gh:SQ.cQvw4UsHHRRlogxbWCB8W-mKD4AraM9y",
        "type": "SequenceLocation"
    }


@pytest.fixture(scope="session")
def braf_v600e(braf_gene_context, braf_600loc):
    """Create BRAF V600E protein test fixture."""
    params = {
        "id": "normalize.variation:BRAF%20V600E",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.ZDdoQdURgO2Daj2NxLj4pcDnjiiAsfbO",
        "variation": {
            "_id": "ga4gh:VA.ZDdoQdURgO2Daj2NxLj4pcDnjiiAsfbO",
            "location": braf_600loc,
            "state": {
                "sequence": "E",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001606",
        "vrs_ref_allele_seq": "V",
        "gene_context": braf_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="session")
def vhl_silent(vhl_gene_context):
    """Create NP_000542.1:p.Pro61 fixture."""
    params = {
        "id": "normalize.variation:NP_000542.1%3Ap.Pro61%3D",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.S1GX6EwJV3exmJAH8MnxS8-S9J4i2Ip_",
        "variation": {
            "_id": "ga4gh:VA.S1GX6EwJV3exmJAH8MnxS8-S9J4i2Ip_",
            "location": {
                "_id": "ga4gh:VSL.zuNGmA02Uq49faqvCIPtwVrF_IJuP4dM",
                "interval": {
                    "end": {"value": 61, "type": "Number"},
                    "start": {"value": 60, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.z-Oa0pZkJ6GHJHOYM7h5mY_umc0SJzTu",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "P",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001017",
        "vrs_ref_allele_seq": "P",
        "gene_context": vhl_gene_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="session")
def protein_insertion(egfr_context):
    """Create test fixture for NP protein insertion."""
    params = {
        "id": "normalize.variation:NP_005219.2%3Ap.Asp770_Asn771insGlyLeu",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.t_WLqe5efVQlBmdbIBgqIeLRu2rSJDJJ",
        "variation": {
            "_id": "ga4gh:VA.t_WLqe5efVQlBmdbIBgqIeLRu2rSJDJJ",
            "location": {
                "_id": "ga4gh:VSL.DJIP1jlxQIro1oC5re8txtH7N8vAvM7A",
                "interval": {
                    "end": {"value": 770, "type": "Number"},
                    "start": {"value": 770, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.vyo55F6mA6n2LgN4cagcdRzOuh38V4mE",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "GL",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001605",
        "gene_context": egfr_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="session")
def protein_deletion_np_range(erbb2_context):
    """Create test fixture for protein deletion using NP accession and
    range for deletion.
    """
    params = {
        "id": "normalize.variation:NP_004439.2%3Ap.Leu755_Thr759del",
        "type": "VariationDescriptor",
        "variation_id": "ga4gh:VA.rFwsfnekdWjwKNmsAw9fZOCGgIvcMnCn",
        "variation": {
            "_id": "ga4gh:VA.rFwsfnekdWjwKNmsAw9fZOCGgIvcMnCn",
            "location": {
                "_id": "ga4gh:VSL.vhpNJ0vsJx3WbnCfwJzxFU-wWyZwvPdL",
                "interval": {
                    "end": {"value": 759, "type": "Number"},
                    "start": {"value": 754, "type": "Number"},
                    "type": "SequenceInterval"
                },
                "sequence_id": "ga4gh:SQ.AF1UFydIo02-bMplonKSfxlWY2q6ze3m",
                "type": "SequenceLocation"
            },
            "state": {
                "sequence": "",
                "type": "LiteralSequenceExpression"
            },
            "type": "Allele"
        },
        "molecule_context": "protein",
        "structural_type": "SO:0001604",
        "vrs_ref_allele_seq": "LRENT",
        "gene_context": erbb2_context
    }
    return VariationDescriptor(**params)


@pytest.fixture(scope="session")
def braf_v600e_genomic_sub():
    """Create test fixture for NC_000007.14:g.140753336A>T"""
    return {
        "_id": "ga4gh:VA.fZiBjQEolbkL0AxjoTZf4SOkFy9J0ebU",
        "location": {
            "_id": "ga4gh:VSL.zga82-TpYiNmBESCfvDvAz9DyvJF98I-",
            "interval": {
                "end": {"value": 140753336, "type": "Number"},
                "start": {"value": 140753335, "type": "Number"},
                "type": "SequenceInterval"
            },
            "sequence_id": "ga4gh:SQ.F-LrLMe1SRpfUZHkQmvkVKFEGaoDeHul",
            "type": "SequenceLocation"
        },
        "state": {
            "sequence": "T",
            "type": "LiteralSequenceExpression"
        },
        "type": "Allele"
    }


@pytest.fixture(scope="session")
def genomic_dup1_seq_loc():
    """Create test fixture containing genomic dup1 sequence location"""
    return {
        "_id": "ga4gh:VSL.G_J9WrfooiONRgjbmGPuCBYbBYFQnYOg",
        "sequence_id": "ga4gh:SQ.Zu7h9AggXxhTaGVsy7h_EZSChSZGcmgX",
        "interval": {
            "type": "SequenceInterval",
            "start": {"value": 49531260, "type": "Number"},
            "end": {"value": 49531262, "type": "Number"},
        },
        "type": "SequenceLocation"
    }


@pytest.fixture(scope="session")
def genomic_dup1_38_vac(genomic_dup1_seq_loc):
    """Create test fixture for absolute copy number dup1 on GRCh38"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.2zTRgNWai56-CSvxw_UerY2ggUz3kJwe",
        "subject": {
            "location": genomic_dup1_seq_loc,
            "reverse_complement": False,
            "type": "DerivedSequenceExpression"
        },
        "copies": {
            "type": "Number",
            "value": 3
        }
    }


@pytest.fixture(scope="session")
def genomic_dup2_seq_loc():
    """Create genomic dup2 sequence location"""
    return {
        "_id": "ga4gh:VSL.4mH68huylkPmu6zyUwH4wiazIYr9cQUX",
        "sequence_id": "ga4gh:SQ.yC_0RBj3fgBlvgyAuycbzdubtLxq-rE0",
        "interval": {
            "type": "SequenceInterval",
            "start": {"value": 2087937, "type": "Number"},
            "end": {"value": 2087948, "type": "Number"},
        },
        "type": "SequenceLocation"
    }


@pytest.fixture(scope="session")
def genomic_dup2_38_vac(genomic_dup2_seq_loc):
    """Create test fixture for absolute copy number dup2 on GRCh38"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.xwy6h-f8s6B3QLiPYKiHZqWjAO7eKe4Y",
        "subject": {
            "location": genomic_dup2_seq_loc,
            "reverse_complement": False,
            "type": "DerivedSequenceExpression"
        },
        "copies": {
            "type": "Number",
            "value": 3
        }
    }


@pytest.fixture(scope="session")
def genomic_del3_dup3_loc():
    """Create genomic del3 dup3 sequence location"""
    return {
        "_id": "ga4gh:VSL.DgEMxYt1AdPe-HZAQbT2AVz5OejICnOj",
        "sequence_id": "ga4gh:SQ.w0WZEvgJF0zf_P4yyTzjjv9oW1z61HHP",
        "interval": {
            "type": "SequenceInterval",
            "start": {
                "min": 31060226,
                "max": 31100350,
                "type": "DefiniteRange"
            },
            "end": {
                "min": 33274279,
                "max": 33417152,
                "type": "DefiniteRange"
            }
        },
        "type": "SequenceLocation"
    }


@pytest.fixture(scope="session")
def genoimc_dup4_loc():
    """Create genoimc dup4 sequence location"""
    return {
        "_id": "ga4gh:VSL.us51izImAQQWr-Hu6Q7HQm-vYvmb-jJo",
        "sequence_id": "ga4gh:SQ.-A1QmD_MatoqxvgVxBLZTONHz9-c7nQo",
        "interval": {
            "type": "SequenceInterval",
            "start": {
                "value": 30417575,
                "comparator": "<=",
                "type": "IndefiniteRange"
            },
            "end": {
                "value": 31394018,
                "comparator": ">=",
                "type": "IndefiniteRange"
            }
        },
        "type": "SequenceLocation"
    }


@pytest.fixture(scope="session")
def genoimc_dup5_loc():
    """Create genoimc dup5 sequence location"""
    return {
        "_id": "ga4gh:VSL.k2FXLyqyS8pbtZxEHCpNd2SHD6iCtH9C",
        "sequence_id": "ga4gh:SQ.w0WZEvgJF0zf_P4yyTzjjv9oW1z61HHP",
        "interval": {
            "type": "SequenceInterval",
            "start": {
                "value": 154021811,
                "comparator": "<=",
                "type": "IndefiniteRange"
            },
            "end": {
                "value": 154092209,
                "type": "Number"
            }
        },
        "type": "SequenceLocation"
    }


@pytest.fixture(scope="session")
def genoimc_dup6_loc():
    """Create genoimc dup6 sequence location"""
    return {
        "_id": "ga4gh:VSL.h0_xXu36uSnPEuLoxvVmTAFQCS1ZFuLN",
        "sequence_id": "ga4gh:SQ.w0WZEvgJF0zf_P4yyTzjjv9oW1z61HHP",
        "interval": {
            "type": "SequenceInterval",
            "start": {
                "value": 154021811,
                "type": "Number"
            },
            "end": {
                "value": 154092209,
                "comparator": ">=",
                "type": "IndefiniteRange"
            }
        },
        "type": "SequenceLocation"
    }


@pytest.fixture(scope="session")
def genomic_del1_seq_loc():
    """Create genomic del1 sequence location"""
    return {
        "_id": "ga4gh:VSL.Yg5B66zErDjK9Lqeaw-kuzAB9w5-uUaS",
        "sequence_id": "ga4gh:SQ.Zu7h9AggXxhTaGVsy7h_EZSChSZGcmgX",
        "interval": {
            "type": "SequenceInterval",
            "start": {"value": 10149810, "type": "Number"},
            "end": {"value": 10149811, "type": "Number"},
        },
        "type": "SequenceLocation"
    }


@pytest.fixture(scope="session")
def genomic_del1_38_vac(genomic_del1_seq_loc):
    """Create test fixture for absolute copy number del1 on GRCh38"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.Gww88B2yhxPAT2nrNWaxo9oPyQ5mPrYW",
        "subject": {
            "location": genomic_del1_seq_loc,
            "reverse_complement": False,
            "type": "DerivedSequenceExpression"
        },
        "copies": {
            "type": "Number",
            "value": 1
        }
    }


@pytest.fixture(scope="session")
def genomic_del2_seq_loc():
    """Create genomic del2 sequence location"""
    return {
        "_id": "ga4gh:VSL.lksYAhEQvP8biy_nxoOJ_Zwu75a_kYtQ",
        "sequence_id": "ga4gh:SQ.Zu7h9AggXxhTaGVsy7h_EZSChSZGcmgX",
        "interval": {
            "type": "SequenceInterval",
            "start": {"value": 10146594, "type": "Number"},
            "end": {"value": 10146613, "type": "Number"},
        },
        "type": "SequenceLocation"
    }


@pytest.fixture(scope="session")
def genomic_del2_38_vac(genomic_del2_seq_loc):
    """Create test fixture for absolute copy number del1 on GRCh38"""
    return {
        "type": "AbsoluteCopyNumber",
        "_id": "ga4gh:VAC.NijRPhc84gKf8xFq4aa3RdOkw6yXAMRt",
        "subject": {
            "location": genomic_del2_seq_loc,
            "reverse_complement": False,
            "type": "DerivedSequenceExpression"
        },
        "copies": {
            "type": "Number",
            "value": 1
        }
    }


@pytest.fixture(scope="session")
def genomic_del4_seq_loc():
    """Create genomic del4 sequence location"""
    return {
        "_id": "ga4gh:VSL.7OJ5EFgu_2C4zPFDUBgn-ziE6BZwsRcv",
        "sequence_id": "ga4gh:SQ.w0WZEvgJF0zf_P4yyTzjjv9oW1z61HHP",
        "interval": {
            "type": "SequenceInterval",
            "start": {
                "value": 31120495,
                "comparator": "<=",
                "type": "IndefiniteRange"
            },
            "end": {
                "value": 33339477,
                "comparator": ">=",
                "type": "IndefiniteRange"
            }
        },
        "type": "SequenceLocation"
    }


@pytest.fixture(scope="session")
def genomic_del5_seq_loc():
    """Create genomic del5 sequence location"""
    return {
        "_id": "ga4gh:VSL.jURzcCBf3kJVx19uuJJtwt78LuBbtfwD",
        "sequence_id": "ga4gh:SQ.w0WZEvgJF0zf_P4yyTzjjv9oW1z61HHP",
        "interval": {
            "type": "SequenceInterval",
            "start": {
                "value": 18575353,
                "comparator": "<=",
                "type": "IndefiniteRange"
            },
            "end": {
                "value": 18653629,
                "type": "Number"
            }
        },
        "type": "SequenceLocation"
    }


@pytest.fixture(scope="session")
def genomic_del6_seq_loc():
    """Create genomic del6 sequence location"""
    return {
        "_id": "ga4gh:VSL.TPwsB5ymsNI7TynTlI8_8CI_NmNrBHUQ",
        "sequence_id": "ga4gh:SQ.0iKlIQk2oZLoeOG9P1riRU6hvL5Ux8TV",
        "interval": {
            "type": "SequenceInterval",
            "start": {
                "value": 133462763,
                "type": "Number"
            },
            "end": {
                "value": 133464858,
                "comparator": ">=",
                "type": "IndefiniteRange"
            }
        },
        "type": "SequenceLocation"
    }


def assertion_checks(normalize_response, test_variation, label, ignore_id=False):
    """Check that normalize_response and test_variation are equal."""
    if not ignore_id:
        assert normalize_response.id == test_variation.id, "id"
    assert normalize_response.label == label
    assert normalize_response.type == test_variation.type, "type"
    assert normalize_response.variation_id == \
           test_variation.variation_id, "variation_id"
    if test_variation.variation.type != "Text":
        if test_variation.variation.id:
            assert normalize_response.variation.id == \
                   test_variation.variation.id, "variation._id"
            if test_variation.variation_id:
                assert normalize_response.variation_id == \
                       normalize_response.variation.id, "variation_id == variation.id"  # noqa: E501
        assert normalize_response.variation == \
               test_variation.variation, "variation"
    else:
        if not ignore_id:
            assert normalize_response.variation.id == \
                   test_variation.variation.id
        assert normalize_response.variation.type == \
               test_variation.variation.type
        assert normalize_response.variation.definition == \
               test_variation.variation.definition
    assert normalize_response.molecule_context == \
           test_variation.molecule_context, "molecule_context"
    assert normalize_response.structural_type == \
           test_variation.structural_type, "structural_type"
    assert normalize_response.vrs_ref_allele_seq == \
           test_variation.vrs_ref_allele_seq, "vrs_ref_allele_seq"

    resp_gene_context = normalize_response.gene_context
    test_variation_context = test_variation.gene_context
    if resp_gene_context:
        if not ignore_id:
            assert resp_gene_context.id == \
                   test_variation_context.id, "gene_context.id"
        assert resp_gene_context.label == \
               test_variation_context.label, "gene_context.label"
        assert resp_gene_context.gene_id ==\
               test_variation_context.gene_id, "gene_context.gene_id"
        assert set(resp_gene_context.xrefs) ==\
               set(test_variation_context.xrefs), "gene_context.xrefs"
        if test_variation_context.alternate_labels:
            assert set(resp_gene_context.alternate_labels) == \
                   set(test_variation_context.alternate_labels), "gene_context.alternate_labels"  # noqa: E501
        assert len(resp_gene_context.extensions) == \
               len(test_variation_context.extensions), "len gene_context.extensions"  # noqa: E501
        for resp_ext in resp_gene_context.extensions:
            for test_var in test_variation_context.extensions:
                if resp_ext.name == test_var.name:
                    if resp_ext.name == "chromosome_location":
                        assert resp_ext.value == test_var.value, \
                            "gene_context.chromosome_location"
                    elif resp_ext.name == "associated_with":
                        assert set(resp_ext.value) == set(test_var.value), \
                            "gene_context.associated_with"
                    else:
                        assert resp_ext.value == test_var.value,\
                            f"gene_context.{resp_ext.name}"
    else:
        assert not test_variation_context



"""Module for Gene Symbol tokenization."""
from typing import Optional

from gene.schemas import MatchType
from gene.query import QueryHandler as GeneQueryHandler

from variation.schemas.token_response_schema import GeneMatchToken, \
    TokenMatchType
from .tokenizer import Tokenizer


class GeneSymbol(Tokenizer):
    """Class for gene symbol tokenization."""

    def __init__(self, gene_normalizer: GeneQueryHandler) -> None:
        """Initialize the gene symbol tokenizer class.

        :param QueryHandler gene_normalizer: Instance to gene normalizer
            QueryHandler
        """
        self.gene_normalizer = gene_normalizer
        self._gene_cache = dict()

    def match(self, input_string: str) -> Optional[GeneMatchToken]:
        """Return token matches from input string."""
        upper_input = input_string.upper()
        if upper_input in self._gene_cache:
            return self._gene_cache[upper_input]

        norm_resp = self.gene_normalizer.normalize(upper_input)

        norm_match_type = norm_resp.match_type
        if norm_match_type == MatchType.CONCEPT_ID:
            match_type = TokenMatchType.ID
        elif norm_match_type == MatchType.SYMBOL:
            match_type = TokenMatchType.SYMBOL
        elif norm_match_type == MatchType.ALIAS:
            match_type = TokenMatchType.ALIAS
        elif norm_match_type == MatchType.PREV_SYMBOL:
            match_type = TokenMatchType.PREVIOUS
        else:
            match_type = TokenMatchType.UNSPECIFIED

        if match_type != TokenMatchType.UNSPECIFIED:
            label = norm_resp.gene_descriptor.label
        else:
            label = upper_input

        if norm_match_type != 0:
            gene_match_token = GeneMatchToken(
                token=label,
                input_string=input_string,
                matched_value=label,
                match_type=match_type
            )
            self._gene_cache[upper_input] = gene_match_token
            return gene_match_token
        else:
            return None



"""Module for testing canonical_spdi_to_categorical_variation"""
import copy

import pytest
from ga4gh.vrsatile.pydantic.vrsatile_models import CanonicalVariation


@pytest.fixture(scope="module")
def spdi1():
    """Create test fixture for NC_000013.11:20189346:GGG:GG"""
    params = {
        "_id": "ga4gh:VCC.6FxWtQdkEyVSMIOnvRj0bOEgvHgN3pRh",
        "complement": False,
        "type": "CanonicalVariation",
        "variation": {
            "_id": "ga4gh:VA.KGopzor-bEw8Ot5sAQQ5o5SVx4o7TuLN",
            "type": "Allele",
            "location": {
                "_id": "ga4gh:VSL.p4e9kMEY9PrKZ1BbNRuFr6n30DkwXWlX",
                "type": "SequenceLocation",
                "sequence_id": "ga4gh:SQ._0wi-qoDrvram155UmcSC-zA5ZK4fpLT",
                "interval": {
                    "type": "SequenceInterval",
                    "start": {
                        "type": "Number",
                        "value": 20189346
                    },
                    "end": {
                        "type": "Number",
                        "value": 20189349
                    }
                }
            },
            "state": {
                "type": "LiteralSequenceExpression",
                "sequence": "GG"
            }
        }
    }
    return CanonicalVariation(**params)


@pytest.fixture(scope="module")
def spdi2(braf_v600e_genomic_sub):
    """Create test fixture for NC_000007.14:140753335:A:T"""
    params = {
        "_id": "ga4gh:VCC.jh-D_L4J74__OdCUI8bn6j4cI1RWZIAA",
        "complement": True,
        "type": "CanonicalVariation",
        "variation": braf_v600e_genomic_sub
    }
    return CanonicalVariation(**params)


def test_canonical_spdi_to_categorical_variation(test_query_handler, spdi1, spdi2):
    """Test that canonical_spdi_to_categorical_variation works correctly"""
    # https://www.ncbi.nlm.nih.gov/clinvar/variation/17014/?new_evidence=true
    q = "NC_000013.11:20189346:GGG:GG"
    resp, w = test_query_handler.canonical_spdi_to_categorical_variation(q)
    assert resp == spdi1
    assert w == []

    q = "NC_000007.14:140753335:A:T"
    resp, w = test_query_handler.canonical_spdi_to_categorical_variation(
        q, complement=True)
    assert resp == spdi2

    resp, w = test_query_handler.canonical_spdi_to_categorical_variation(
        q, complement=False)
    cpy_spdi2 = copy.deepcopy(spdi2)
    cpy_spdi2.complement = False
    cpy_spdi2.id = "ga4gh:VCC.W0r_NF_ecKXjgvTwcMNkyVS1pB_CXMj9"
    assert resp == cpy_spdi2


def test_invalid(test_query_handler):
    """Test that invalid queries return the correct response"""
    resp, w = test_query_handler.canonical_spdi_to_categorical_variation(
        "NC_000013.11:201845654659346:GGG:GG")
    assert resp is None
    assert w == ["start out of range (201845654659346)"]

    resp, w = test_query_handler.canonical_spdi_to_categorical_variation(
        "NC_000013.11:2018459346:GGG:GG")
    assert resp is None
    assert w == ["Position, 2018459346, does not exist on NC_000013.11"]

    resp, w = test_query_handler.canonical_spdi_to_categorical_variation(
        "NC_000013.1:20189346:GGG:GG")
    assert resp is None
    assert w == ["vrs-python translator raised error: seqrepo could not translate "
                 "identifier 'refseq:NC_000013.1'"]

    resp, w = test_query_handler.canonical_spdi_to_categorical_variation(
        "NP_004324.2:p.Val600Glu")
    assert resp is None
    assert w == ["vrs-python translator raised error: Unable to parse data as "
                 "spdi variation"]

    resp, w = test_query_handler.canonical_spdi_to_categorical_variation(
        "NC_000013.11:20189346:GCG:GG")
    assert resp is None
    assert w ==\
           ["Expected to find reference sequence GCG but found GGG on NC_000013.11"]



"""Module for Coding DNA Deletion Translation."""
from variation.translators.translator import Translator
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import CodingDNADeletionToken, Token


class CodingDNADeletion(Translator):
    """The Coding DNA Deletion Translator class."""

    def can_translate(self, type: ClassificationType) -> bool:
        """Return if classification type is Coding DNA Deletion."""
        return type == ClassificationType.CODING_DNA_DELETION

    def is_token_instance(self, token: Token) -> bool:
        """Return if the token is an Coding DNA Deletion token instance."""
        return isinstance(token, CodingDNADeletionToken)



"""Module containing schemas used in HGVS To Copy Number endpoints"""
from enum import Enum
from typing import Type, Any, Dict, Union

from ga4gh.vrsatile.pydantic.vrs_models import RelativeCopyClass, AbsoluteCopyNumber, \
    RelativeCopyNumber, Text
from pydantic import StrictStr

from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.normalize_response_schema import ServiceResponse


VALID_CLASSIFICATION_TYPES = [
    ClassificationType.GENOMIC_DUPLICATION,
    ClassificationType.GENOMIC_DELETION,
    ClassificationType.GENOMIC_DELETION_RANGE,
    ClassificationType.GENOMIC_UNCERTAIN_DELETION
]


class CopyNumberType(str, Enum):
    """Define copy number types"""

    RELATIVE = "relative_copy_number"
    ABSOLUTE = "absolute_copy_number"


VALID_RELATIVE_COPY_CLASS = [rcc.value for
                             rcc in RelativeCopyClass.__members__.values()]


class HgvsToAbsoluteCopyNumberService(ServiceResponse):
    """A response for translating HGVS to absolute copy number."""

    hgvs_expr: StrictStr
    absolute_copy_number: Union[AbsoluteCopyNumber, Text]

    class Config:
        """Configure model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["HgvsToAbsoluteCopyNumberService"]) -> None:
            """Configure OpenAPI schema."""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "hgvs_expr": "NC_000003.12:g.49531262dup",
                "absolute_copy_number": {
                    "_id": "ga4gh:VAC.2zTRgNWai56-CSvxw_UerY2ggUz3kJwe",
                    "type": "AbsoluteCopyNumber",
                    "subject": {
                        "type": "DerivedSequenceExpression",
                        "location": {
                            "_id": "ga4gh:VSL.G_J9WrfooiONRgjbmGPuCBYbBYFQnYOg",
                            "type": "SequenceLocation",
                            "sequence_id": "ga4gh:SQ.Zu7h9AggXxhTaGVsy7h_EZSChSZGcmgX",
                            "interval": {
                                "type": "SequenceInterval",
                                "start": {"type": "Number", "value": 49531260},
                                "end": {"type": "Number", "value": 49531262}
                            }
                        },
                        "reverse_complement": False
                    },
                    "copies": {"type": "Number", "value": 3}
                },
                "service_meta_": {
                    "name": "variation-normalizer",
                    "version": "0.2.17",
                    "response_datetime": "2022-01-26T22:23:41.821673",
                    "url": "https://github.com/cancervariants/variation-normalization"
                }
            }


class HgvsToRelativeCopyNumberService(ServiceResponse):
    """A response for translating HGVS to relative copy number."""

    hgvs_expr: StrictStr
    relative_copy_number: Union[RelativeCopyNumber, Text]

    class Config:
        """Configure model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["HgvsToRelativeCopyNumberService"]) -> None:
            """Configure OpenAPI schema."""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "hgvs_expr": "NC_000003.12:g.49531262dup",
                "relative_copy_number": {
                    "_id": "ga4gh:VRC.XiXamTGYJ43rc8xheleMKcjxEBOFp82l",
                    "type": "RelativeCopyNumber",
                    "subject": {
                        "type": "DerivedSequenceExpression",
                        "location": {
                            "_id": "ga4gh:VSL.G_J9WrfooiONRgjbmGPuCBYbBYFQnYOg",
                            "type": "SequenceLocation",
                            "sequence_id": "ga4gh:SQ.Zu7h9AggXxhTaGVsy7h_EZSChSZGcmgX",
                            "interval": {
                                "type": "SequenceInterval",
                                "start": {"type": "Number", "value": 49531260},
                                "end": {"type": "Number", "value": 49531262}
                            }
                        },
                        "reverse_complement": False
                    },
                    "relative_copy_class": "complete loss"
                },
                "service_meta_": {
                    "name": "variation-normalizer",
                    "version": "0.2.17",
                    "response_datetime": "2022-01-26T22:23:41.821673",
                    "url": "https://github.com/cancervariants/variation-normalization"
                }
            }



"""Module for testing protein insertion Translator."""
import unittest

from variation.classifiers import ProteinInsertionClassifier
from variation.translators import ProteinInsertion
from variation.validators import ProteinInsertion as AAI_V
from .translator_base import TranslatorBase


class TestProteinInsertionTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the protein insertion Translator."""

    def classifier_instance(self):
        """Return protein insertion instance."""
        return ProteinInsertionClassifier()

    def validator_instance(self):
        """Return protein insertion instance."""
        return AAI_V(*self.aa_params)

    def translator_instance(self):
        """Return protein insertion instance."""
        return ProteinInsertion()

    def fixture_name(self):
        """Return the fixture name for protein insertion."""
        return "protein_insertion"



"""The module for Protein Deletion Validation."""
import logging
from typing import Dict, List, Optional

from ga4gh.vrsatile.pydantic.vrs_models import RelativeCopyClass
from gene.query import QueryHandler as GeneQueryHandler
from ga4gh.vrs.dataproxy import SeqRepoDataProxy
from ga4gh.vrs.extras.translator import Translator
from uta_tools.data_sources import SeqRepoAccess, TranscriptMappings, UTADatabase, \
    MANETranscript

from variation.schemas.classification_response_schema import \
    Classification, ClassificationType
from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum
from variation.schemas.app_schemas import Endpoint
from variation.schemas.token_response_schema import Token
from variation.validators.validator import Validator
from variation.schemas.token_response_schema import GeneMatchToken
from variation.tokenizers import GeneSymbol
from variation.tokenizers.caches import AminoAcidCache
from variation.vrs import VRS
from .protein_base import ProteinBase

logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)


class ProteinDeletion(Validator):
    """The Protein Deletion Validator class."""

    def __init__(self, seq_repo_access: SeqRepoAccess,
                 transcript_mappings: TranscriptMappings,
                 gene_symbol: GeneSymbol,
                 mane_transcript: MANETranscript,
                 uta: UTADatabase, dp: SeqRepoDataProxy, tlr: Translator,
                 gene_normalizer: GeneQueryHandler, vrs: VRS,
                 amino_acid_cache: AminoAcidCache) \
            -> None:
        """Initialize the validator.

        :param SeqRepoAccess seq_repo_access: Access to SeqRepo data
        :param TranscriptMappings transcript_mappings: Access to transcript
            mappings
        :param GeneSymbol gene_symbol: Gene symbol tokenizer
        :param MANETranscript mane_transcript: Access MANE Transcript
            information
        :param UTADatabase uta: Access to UTA queries
        :param GeneQueryHandler gene_normalizer: Access to gene-normalizer
        :param VRS vrs: Class for creating VRS objects
        :param amino_acid_cache: Amino Acid codes and conversions
        """
        super().__init__(
            seq_repo_access, transcript_mappings, gene_symbol, mane_transcript,
            uta, dp, tlr, gene_normalizer, vrs
        )
        self._amino_acid_cache = amino_acid_cache
        self.protein_base = ProteinBase(seq_repo_access, amino_acid_cache)
        self.mane_transcript = mane_transcript

    async def get_transcripts(self, gene_tokens: List, classification: Classification,
                              errors: List) -> Optional[List[str]]:
        """Get transcript accessions for a given classification.

        :param List gene_tokens: A list of gene tokens
        :param Classification classification: A classification for a list of
            tokens
        :param List errors: List of errors
        :return: List of transcript accessions
        """
        return self.get_protein_transcripts(gene_tokens, errors)

    async def get_valid_invalid_results(
        self, classification_tokens: List, transcripts: List,
        classification: Classification, results: List, gene_tokens: List,
        mane_data_found: Dict, is_identifier: bool,
        hgvs_dup_del_mode: HGVSDupDelModeEnum,
        endpoint_name: Optional[Endpoint] = None,
        baseline_copies: Optional[int] = None,
        relative_copy_class: Optional[RelativeCopyClass] = None,
        do_liftover: bool = False
    ) -> None:
        """Add validation result objects to a list of results.

        :param List classification_tokens: A list of classification Tokens
        :param List transcripts: A list of transcript accessions
        :param Classification classification: A classification for a list of
            tokens
        :param List results: Stores validation result objects
        :param List gene_tokens: List of GeneMatchTokens for a classification
        :param Dict mane_data_found: MANE Transcript information found
        :param bool is_identifier: `True` if identifier is given for exact
            location. `False` otherwise.
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Optional[Endpoint] endpoint_name: Then name of the endpoint being used
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        """
        valid_alleles = list()
        for s in classification_tokens:
            for t in transcripts:
                errors = list()
                t = self.get_accession(t, classification)
                allele = self.vrs.to_vrs_allele(
                    t, s.start_pos_del, s.end_pos_del, s.coordinate_type,
                    s.alt_type, errors)

                if not errors:
                    # Check ref start/end protein matches expected
                    self.protein_base.check_ref_aa(
                        t, s.start_aa_del, s.start_pos_del, errors
                    )

                    if not errors and s.end_aa_del:
                        self.protein_base.check_ref_aa(
                            t, s.end_aa_del, s.end_pos_del, errors
                        )

                if not errors and endpoint_name == Endpoint.NORMALIZE:
                    mane = await self.mane_transcript.get_mane_transcript(
                        t, s.start_pos_del, s.coordinate_type, end_pos=s.end_pos_del,
                        try_longest_compatible=True)
                    self.add_mane_data(mane, mane_data_found, s.coordinate_type,
                                       s.alt_type, s)

                self.add_validation_result(allele, valid_alleles, results,
                                           classification, s, t, gene_tokens, errors)

                if is_identifier:
                    break

        if endpoint_name == Endpoint.NORMALIZE:
            self.add_mane_to_validation_results(mane_data_found, valid_alleles, results,
                                                classification, gene_tokens)

    def get_gene_tokens(self, classification: Classification) -> List[GeneMatchToken]:
        """Return gene tokens for a classification.

        :param Classification classification: The classification for tokens
        :return: A list of Gene Match Tokens in the classification
        """
        return self.get_protein_gene_symbol_tokens(classification)

    def variation_name(self) -> str:
        """Return the variation name."""
        return "protein deletion"

    def is_token_instance(self, t: Token) -> bool:
        """Check that token is Protein DelIns."""
        return t.token_type == "ProteinDeletion"

    def validates_classification_type(
            self, classification_type: ClassificationType) -> bool:
        """Return whether or not the classification type is
        Protein Deletion.
        """
        return classification_type == ClassificationType.PROTEIN_DELETION



"""A module for testing validator classes."""
import yaml
from ga4gh.vrs.dataproxy import SeqRepoDataProxy
from ga4gh.vrs.extras.translator import Translator
from gene.query import QueryHandler as GeneQueryHandler
from uta_tools.data_sources import TranscriptMappings, SeqRepoAccess, \
    MANETranscriptMappings, UTADatabase, MANETranscript

from tests import PROJECT_ROOT
from variation.schemas.app_schemas import Endpoint
from variation.vrs import VRS
from variation.tokenizers import Tokenize, GeneSymbol
from variation.tokenizers.caches import AminoAcidCache


class ValidatorBase:
    """The validator base class."""

    @classmethod
    def setUpClass(cls):
        """Set up the test cases."""
        with open(f"{PROJECT_ROOT}/tests/fixtures/validators.yml") as stream:
            cls.all_fixtures = yaml.safe_load(stream)
        amino_acid_cache = AminoAcidCache()
        gene_normalizer = GeneQueryHandler()
        gene_symbol = GeneSymbol(gene_normalizer)
        cls.tokenizer = Tokenize(amino_acid_cache, gene_symbol)

        seqrepo_access = SeqRepoAccess()
        transcript_mappings = TranscriptMappings()
        uta = UTADatabase()
        dp = SeqRepoDataProxy(seqrepo_access.seqrepo_client)
        tlr = Translator(data_proxy=dp)
        mane_transcript = MANETranscript(
            seqrepo_access, transcript_mappings, MANETranscriptMappings(), uta)
        vrs = VRS(dp, seqrepo_access)

        cls.aa_params = [
            seqrepo_access, transcript_mappings, gene_symbol,
            mane_transcript, uta, dp, tlr, gene_normalizer,
            vrs, amino_acid_cache
        ]
        cls.params = cls.aa_params[:-1]

    def classifier_instance(self):
        """Check that the classifier_instance method is implemented."""
        raise NotImplementedError()

    def validator_instance(self, **kwargs):
        """Check that the validator_instance method is implemented."""
        raise NotImplementedError()

    def fixture_name(self):
        """Check that the fixture_name method is implemented."""
        raise NotImplementedError()

    def set_up(self):
        """Initialize fixtures, classifier + validator"""
        self.fixtures = self.all_fixtures.get(
            self.fixture_name(),
            {"should_match": [], "should_not_match": []}
        )
        self.classifier = self.classifier_instance()
        self.validator = self.validator_instance()

    async def test_matches(self):
        """Test that validator matches correctly."""
        self.set_up()
        for x in self.fixtures["should_match"]:
            tokens = self.tokenizer.perform(x["query"], [])
            classification = self.classifier.match(tokens)
            validation_results = await self.validator.validate(
                classification, hgvs_dup_del_mode="default",
                endpoint_name=Endpoint.TO_VRS
            )
            is_valid = False
            for vr in validation_results:
                if vr.is_valid:
                    is_valid = True
                    break
            self.assertTrue(is_valid, msg=x)
            self.assertIsNotNone(validation_results, msg=x)

    async def test_not_matches(self):
        """Test that validator matches correctly."""
        self.set_up()
        for x in self.fixtures["should_not_match"]:
            tokens = self.tokenizer.perform(x["query"], [])
            classification = self.classifier.match(tokens)
            validation_results = await self.validator.validate(
                classification, endpoint_name=Endpoint.TO_VRS
            )
            is_valid = False
            for vr in validation_results:
                if vr.is_valid:
                    is_valid = True
                    break
            self.assertFalse(is_valid, msg=x)



"""A module for testing classifier classes."""
import yaml
from gene.query import QueryHandler as GeneQueryHandler

from variation.tokenizers import Tokenize
from tests import PROJECT_ROOT
from variation.tokenizers.caches import AminoAcidCache
from variation.tokenizers import GeneSymbol


class ClassifierBase:
    """The classifier base class."""

    def setUp(self):
        """Set up the test cases."""
        with open(f"{PROJECT_ROOT}/tests/fixtures/classifiers.yml") as stream:
            self.all_fixtures = yaml.safe_load(stream)
        self.fixtures = self.all_fixtures.get(
            self.fixture_name(),
            {"should_match": [], "should_not_match": []}
        )
        self.classifier = self.classifier_instance()
        self.tokenizer = Tokenize(AminoAcidCache(),
                                  GeneSymbol(GeneQueryHandler()))

    def classifier_instance(self):
        """Check that the classifier_instance method is implemented."""
        raise NotImplementedError()

    def fixture_name(self):
        """Check that the fixture_name method is implemented."""
        raise NotImplementedError()

    def test_matches(self):
        """Test that classifier matches correctly."""
        for x in self.fixtures["should_match"]:
            tokens = self.tokenizer.perform(x["query"], [])
            classification = self.classifier.match(tokens)
            self.assertIsNotNone(classification, msg=x)
            self.assertEqual(x["confidence"],
                             str(classification.confidence),
                             msg=x)

    def test_not_matches(self):
        """Test that classifier matches correctly."""
        for x in self.fixtures["should_not_match"]:
            tokens = self.tokenizer.perform(x["query"], [])
            classification = self.classifier.match(tokens)
            self.assertIsNone(classification, msg=x)



"""Module for tokenizing gene pair."""
from typing import Optional

from gene.query import QueryHandler as GeneQueryHandler

from variation.schemas.token_response_schema import GenePairMatchToken, \
    TokenMatchType
from .tokenizer import Tokenizer
from .gene_symbol import GeneSymbol


class GenePair(Tokenizer):
    """The gene pair class."""

    def __init__(self) -> None:
        """Initialize the Gene Pair class."""
        self.__gene_matcher = GeneSymbol(GeneQueryHandler())

    def match(self, input_string: str) -> Optional[GenePairMatchToken]:
        """Return tokens that match the input string."""
        potential_genes = input_string.split("-")
        if len(potential_genes) == 2:
            first_match = self.__gene_matcher.match(potential_genes[0].upper())
            second_match = \
                self.__gene_matcher.match(potential_genes[1].upper())
            if first_match and second_match:
                return GenePairMatchToken(
                    token=f"{first_match.token}-{second_match.token}",
                    match_type=TokenMatchType.UNSPECIFIED,
                    input_string=input_string,
                    left_gene_token=first_match,
                    right_gene_token=second_match
                )
        return None



"""A module for testing Protein Insertion Tokenization."""
import unittest

from variation.tokenizers import ProteinInsertion
from variation.tokenizers.caches import AminoAcidCache, NucleotideCache
from .tokenizer_base import TokenizerBase


class TestProteinInsertionTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Protein Insertion Tokenization."""

    def tokenizer_instance(self):
        """Return Protein Insertion instance."""
        return ProteinInsertion(AminoAcidCache(), NucleotideCache())

    def token_type(self):
        """Return protein insertion token type."""
        return "ProteinInsertion"

    def fixture_name(self):
        """Return the fixture name for protein insertion."""
        return "protein_insertion"



"""Module for testing Genomic Insertion Classifier."""
import unittest

from variation.classifiers import GenomicInsertionClassifier
from .classifier_base import ClassifierBase


class TestGenomicInsertionClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Genomic Insertion Classifier."""

    def classifier_instance(self):
        """Return GenomicInsertionClassifier instance."""
        return GenomicInsertionClassifier()

    def fixture_name(self):
        """Return GenomicInsertionClassifier fixture name."""
        return "genomic_insertion"



"""A module for Single Nucleotide Variation Tokenization Base Class."""
import re
from abc import abstractmethod
from typing import List, Optional, Dict

from variation.schemas.token_response_schema import \
    SingleNucleotideVariation, TokenMatchType
from variation.tokenizers.caches import NucleotideCache
from .tokenizer import Tokenizer


class SingleNucleotideVariationBase(Tokenizer):
    """Class for tokenizing Single Nucleotide Variations."""

    def __init__(self) -> None:
        """Initialize the Single Nucleotide Variation Base Class."""
        self.splitter = re.compile(r"(\d+)")
        self.sub = None
        self.nucleotide_cache = NucleotideCache()

    def match(self, input_string: str) -> Optional[SingleNucleotideVariation]:
        """Return a SingleNucleotideVariationToken match if one exists.

        :param str input_string: The input string to match
        :return: A SingleNucleotideVariationToken if a match one exists.
        """
        if input_string is None:
            return None

        input_string = str(input_string).lower()
        self.sub = {
            "ref_nucleotide": None,
            "position": None,
            "new_nucleotide": None,
            "coordinate_type": None
        }

        # TODO: Need to add m., and n.
        if "c." not in input_string and "g." not in input_string:
            return None

        sub_parts = self.splitter.split(input_string)
        self._get_sub(sub_parts)
        if None not in self.sub.values():
            params = {
                "token": input_string,
                "input_string": input_string,
                "match_type": TokenMatchType.UNSPECIFIED.value,
                "position": self.sub["position"],
                "ref_nucleotide": self.sub["ref_nucleotide"],
                "new_nucleotide": self.sub["new_nucleotide"]
            }
            return self.return_token(params)
        else:
            if self.sub["position"] is not None and \
                    self.sub["new_nucleotide"] == "=" and \
                    self.sub["coordinate_type"] is not None:
                params = {
                    "token": input_string,
                    "input_string": input_string,
                    "match_type": TokenMatchType.UNSPECIFIED.value,
                    "position": self.sub["position"],
                    "new_nucleotide": self.sub["new_nucleotide"]
                }
                return self.return_token(params)

        return None

    def _get_sub(self, sub_parts: List[str]) -> None:
        """Get parts for SNV.

        :param List[str] sub_parts: Split input string
        """
        if len(sub_parts) == 3:
            if "(" in sub_parts[0] and ")" in sub_parts[2]:
                sub_parts[0] = sub_parts[0].split("(")[-1]
                sub_parts[2] = sub_parts[2].split(")")[0]

            if sub_parts[1].isdigit():
                if ">" in sub_parts[2]:
                    # Substitution
                    ref_nuc, new_nuc = sub_parts[2].split(">")
                    nucleotides = self.nucleotide_cache.nucleotides.keys()
                    if ref_nuc.upper() in nucleotides \
                            and new_nuc.upper() in nucleotides:  # noqa: E501
                        self._set_sub(ref_nuc, sub_parts[1], new_nuc,
                                      sub_parts[0].split(".")[0])
                elif sub_parts[2] == "=":
                    # Silent Mutation
                    self._set_sub(None, sub_parts[1], sub_parts[2],
                                  sub_parts[0].split(".")[0])

    def _set_sub(self, ref_nucleotide: str, position: int, new_nucleotide: str,
                 coordinate_type: str) -> None:
        """Initialize substitution.

        :param str ref_nucleotide: Nucleotide at reference position
        :param str position: The position nucleotide substituted
        :param str new_nucleotide: The substituted nucleotide
        :param str coordinate_type: The coordinate type for the variation
        """
        if ref_nucleotide:
            self.sub["ref_nucleotide"] = ref_nucleotide.upper()
        self.sub["position"] = int(position)
        self.sub["new_nucleotide"] = new_nucleotide.upper()
        if coordinate_type in ["c", "g"]:
            self.sub["coordinate_type"] = coordinate_type

    @abstractmethod
    def return_token(self,
                     params: Dict[str, str]) -> Optional[SingleNucleotideVariation]:
        """Return token instance."""
        raise NotImplementedError



"""A module for the Protein DelIns Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class ProteinDelInsClassifier(SetBasedClassifier):
    """The Protein DelIns Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Protein DelIns classification type."""
        return ClassificationType.PROTEIN_DELINS

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["ProteinDelIns"],
            ["GeneSymbol", "ProteinSubstitution", "ProteinDelIns"],
            ["ProteinDelIns", "GeneSymbol"],
            ["GeneSymbol", "ProteinDelIns"],
            ["HGVS", "ProteinDelIns"],
            ["ReferenceSequence", "ProteinDelIns"],
            ["LocusReferenceGenomic", "ProteinDelIns"]
        ]



"""The module for Polypeptide Sequence Variation Validation."""
import logging
from typing import List, Optional, Dict

from gene.query import QueryHandler as GeneQueryHandler
from ga4gh.vrs.dataproxy import SeqRepoDataProxy
from ga4gh.vrs.extras.translator import Translator
from ga4gh.vrsatile.pydantic.vrs_models import RelativeCopyClass
from uta_tools.data_sources import SeqRepoAccess, TranscriptMappings, UTADatabase, \
    MANETranscript

from variation.schemas.app_schemas import Endpoint
from variation.schemas.token_response_schema import GeneMatchToken
from variation.tokenizers import GeneSymbol
from variation.tokenizers.caches import AminoAcidCache
from variation.schemas.classification_response_schema import Classification
from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum
from variation.vrs import VRS
from .validator import Validator
from .protein_base import ProteinBase

logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)


class PolypeptideSequenceVariationBase(Validator):
    """The Polypeptide Sequence Variation Validator Base class."""

    def __init__(self, seq_repo_access: SeqRepoAccess,
                 transcript_mappings: TranscriptMappings,
                 gene_symbol: GeneSymbol,
                 mane_transcript: MANETranscript,
                 uta: UTADatabase, dp: SeqRepoDataProxy, tlr: Translator,
                 gene_normalizer: GeneQueryHandler, vrs: VRS,
                 amino_acid_cache: AminoAcidCache) -> None:
        """Initialize the validator.

        :param SeqRepoAccess seq_repo_access: Access to SeqRepo data
        :param TranscriptMappings transcript_mappings: Access to transcript
            mappings
        :param GeneSymbol gene_symbol: Gene symbol tokenizer
        :param MANETranscript mane_transcript: Access MANE Transcript
            information
        :param UTADatabase uta: Access to UTA queries
        :param GeneQueryHandler gene_normalizer: Access to gene-normalizer
        :param VRS vrs: Class for creating VRS objects
        :param amino_acid_cache: Amino Acid codes and conversions
        """
        super().__init__(
            seq_repo_access, transcript_mappings, gene_symbol, mane_transcript,
            uta, dp, tlr, gene_normalizer, vrs
        )
        self._amino_acid_cache = amino_acid_cache
        self.protein_base = ProteinBase(seq_repo_access, amino_acid_cache)
        self.mane_transcript = mane_transcript

    async def get_transcripts(self, gene_tokens: List, classification: Classification,
                              errors: List) -> Optional[List[str]]:
        """Get transcript accessions for a given classification.

        :param List gene_tokens: A list of gene tokens
        :param Classification classification: A classification for a list of
            tokens
        :param List errors: List of errors
        :return: List of transcript accessions
        """
        return self.get_protein_transcripts(gene_tokens, errors)

    async def get_valid_invalid_results(
        self, classification_tokens: List, transcripts: List,
        classification: Classification, results: List, gene_tokens: List,
        mane_data_found: Dict, is_identifier: bool,
        hgvs_dup_del_mode: HGVSDupDelModeEnum,
        endpoint_name: Optional[Endpoint] = None,
        baseline_copies: Optional[int] = None,
        relative_copy_class: Optional[RelativeCopyClass] = None,
        do_liftover: bool = False
    ) -> None:
        """Add validation result objects to a list of results.

        :param List classification_tokens: A list of classification Tokens
        :param List transcripts: A list of transcript accessions
        :param Classification classification: A classification for a list of
            tokens
        :param List results: Stores validation result objects
        :param List gene_tokens: List of GeneMatchTokens for a classification
        :param Dict mane_data_found: MANE Transcript information found
        :param bool is_identifier: `True` if identifier is given for exact
            location. `False` otherwise.
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Optional[Endpoint] endpoint_name: Then name of the endpoint being used
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        """
        valid_alleles = list()
        for s in classification_tokens:
            for t in transcripts:
                errors = list()

                t = self.get_accession(t, classification)
                allele = self.vrs.to_vrs_allele(
                    t, s.position, s.position, s.coordinate_type,
                    s.alt_type, errors, alt=s.alt_protein)

                if not errors:
                    self.protein_base.check_ref_aa(t, s.ref_protein, s.position, errors)

                if not errors and endpoint_name == Endpoint.NORMALIZE:
                    mane = await self.mane_transcript.get_mane_transcript(
                        t, s.position, s.coordinate_type, end_pos=s.position,
                        ref=s.ref_protein, try_longest_compatible=True
                    )

                    self.add_mane_data(mane, mane_data_found, s.coordinate_type,
                                       s.alt_type, s, alt=s.alt_protein)

                self.add_validation_result(allele, valid_alleles, results,
                                           classification, s, t, gene_tokens, errors)
                if is_identifier:
                    break

        if endpoint_name == Endpoint.NORMALIZE:
            self.add_mane_to_validation_results(mane_data_found, valid_alleles, results,
                                                classification, gene_tokens)

    def get_gene_tokens(self, classification: Classification) -> List[GeneMatchToken]:
        """Return gene tokens for a classification.

        :param Classification classification: The classification for tokens
        :return: A list of Gene Match Tokens in the classification
        """
        return self.get_protein_gene_symbol_tokens(classification)



"""A module for testing the Gene Pair Tokenizer."""
import unittest

from gene.query import QueryHandler as GeneQueryHandler

from variation.tokenizers import GeneSymbol
from .tokenizer_base import TokenizerBase


class TestGenePairTokenizer(TokenizerBase, unittest.TestCase):
    """The Gene Pair Tokenizer class."""

    def tokenizer_instance(self):
        """Return the Gene Pair tokenizer instance."""
        return GeneSymbol(GeneQueryHandler())

    def token_type(self):
        """Return the Gene Pair token type."""
        return "GeneSymbol"

    def fixture_name(self):
        """Return the fixture name for Gene Pair."""
        return "gene"



"""Module for testing Coding DNA Insertion Classifier."""
import unittest

from variation.classifiers import CodingDNAInsertionClassifier
from .classifier_base import ClassifierBase


class TestCodingDNAInsertionClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Coding DNA Insertion Classifier."""

    def classifier_instance(self):
        """Return CodingDNAInsertionClassifier instance."""
        return CodingDNAInsertionClassifier()

    def fixture_name(self):
        """Return CodingDNAInsertionClassifier fixture name."""
        return "coding_dna_insertion"



"""Module for HGVS tokenization."""
from typing import Optional

from hgvs.parser import Parser
from hgvs.exceptions import HGVSParseError, HGVSInvalidVariantError
from hgvs.validator import IntrinsicValidator

from variation.tokenizers.reference_sequence import REFSEQ_PREFIXES
from variation.schemas.token_response_schema import Token, TokenMatchType, \
    Nomenclature
from .tokenizer import Tokenizer
from .locus_reference_genomic import LocusReferenceGenomic


class HGVS(Tokenizer):
    """The HGVS tokenizer class."""

    def __init__(self) -> None:
        """Initialize the HGVS tokenizer class."""
        self.parser = Parser()
        self.validator = IntrinsicValidator()

    def match(self, input_string: str) -> Optional[Token]:
        """Return token matches from input string."""
        valid_prefix = False
        for prefix in REFSEQ_PREFIXES:
            if input_string.upper().startswith(prefix):
                valid_prefix = True
                break

        if not valid_prefix:
            return None

        try:
            variation = self.parser.parse_hgvs_variant(input_string)
            self.validator.validate(variation)
            if input_string[:3].upper().startswith("LRG"):
                lrg_token = LocusReferenceGenomic().match(
                    input_string.split(":")[0]
                )
                lrg_token.input_string = input_string
                return lrg_token
            else:
                return Token(
                    token=input_string,
                    token_type="HGVS",
                    input_string=input_string,
                    match_type=TokenMatchType.UNSPECIFIED,
                    nomenclature=Nomenclature.HGVS
                )
        except (HGVSParseError, HGVSInvalidVariantError, AttributeError):
            return None



"""A module for Coding DNA Insertion Tokenization."""
from typing import Dict, Optional

from variation.schemas.token_response_schema import CodingDNAInsertionToken
from variation.tokenizers.insertion_base import InsertionBase


class CodingDNAInsertion(InsertionBase):
    """Class for tokenizing Insertion at the coding dna reference sequence."""

    def return_token(self, params: Dict) -> Optional[CodingDNAInsertionToken]:
        """Return coding DNA Insertion token."""
        if self.parts["coordinate_type"] == "c":
            return CodingDNAInsertionToken(**params)



"""The base class for Duplication and Deletion Validation."""
import logging
from typing import Optional, List, Dict, Tuple

from ga4gh.vrsatile.pydantic.vrs_models import RelativeCopyClass
from ga4gh.vrs.dataproxy import SeqRepoDataProxy
from ga4gh.vrs.extras.translator import Translator
from gene.query import QueryHandler as GeneQueryHandler
from uta_tools.data_sources import SeqRepoAccess, TranscriptMappings, UTADatabase, \
    MANETranscript

from variation.schemas.classification_response_schema import Classification, \
    ClassificationType
from variation.schemas.app_schemas import Endpoint
from variation.schemas.token_response_schema import Token, GeneMatchToken
from variation.validators.validator import Validator
from variation.hgvs_dup_del_mode import HGVSDupDelMode
from variation.tokenizers import GeneSymbol
from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum
from variation.vrs import VRS

logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)


class DuplicationDeletionBase(Validator):
    """The Deletion Validator Base class."""

    def __init__(self, seq_repo_access: SeqRepoAccess,
                 transcript_mappings: TranscriptMappings,
                 gene_symbol: GeneSymbol,
                 mane_transcript: MANETranscript,
                 uta: UTADatabase, dp: SeqRepoDataProxy, tlr: Translator,
                 gene_normalizer: GeneQueryHandler, vrs: VRS) -> None:
        """Initialize the Deletion Base validator.

        :param SeqRepoAccess seq_repo_access: Access to SeqRepo data
        :param TranscriptMappings transcript_mappings: Access to transcript
            mappings
        :param GeneSymbol gene_symbol: Gene symbol tokenizer
        :param MANETranscript mane_transcript: Access MANE Transcript
            information
        :param UTADatabase uta: Access to UTA queries
        :param GeneQueryHandler gene_normalizer: Access to gene-normalizer
        :param VRS vrs: Class for creating VRS objects
        """
        super().__init__(
            seq_repo_access, transcript_mappings, gene_symbol, mane_transcript,
            uta, dp, tlr, gene_normalizer, vrs
        )
        self.hgvs_dup_del_mode = HGVSDupDelMode(seq_repo_access)

    def is_token_instance(self, t: Token) -> bool:
        """Check to see if token is instance of a token type.

        :param Token t: Classification token to find type of
        :return: `True` if token is instance of class token. `False` otherwise.
        """
        raise NotImplementedError

    def variation_name(self) -> str:
        """Return the variation name.

        :return: variation class name
        """
        raise NotImplementedError

    def get_gene_tokens(self, classification: Classification) -> List[GeneMatchToken]:
        """Return a list of gene tokens for a classification.

        :param Classification classification: Classification for a list of
            tokens
        :return: A list of gene tokens for the classification
        """
        raise NotImplementedError

    async def get_transcripts(self, gene_tokens: List, classification: Classification,
                              errors: List) -> Optional[List[str]]:
        """Get transcript accessions for a given classification.

        :param List gene_tokens: A list of gene tokens
        :param Classification classification: A classification for a list of
            tokens
        :param List errors: List of errors
        :return: List of transcript accessions
        """
        raise NotImplementedError

    def validates_classification_type(
            self, classification_type: ClassificationType) -> bool:
        """Check that classification type can be validated by validator.

        :param ClassificationType classification_type: The type of variation
        :return: `True` if classification_type matches validator's
            classification type. `False` otherwise.
        """
        raise NotImplementedError

    async def get_valid_invalid_results(
        self, classification_tokens: List, transcripts: List,
        classification: Classification, results: List, gene_tokens: List,
        mane_data_found: Dict, is_identifier: bool,
        hgvs_dup_del_mode: HGVSDupDelModeEnum,
        endpoint_name: Optional[Endpoint] = None,
        baseline_copies: Optional[int] = None,
        relative_copy_class: Optional[RelativeCopyClass] = None,
        do_liftover: bool = False
    ) -> None:
        """Add validation result objects to a list of results.

        :param List classification_tokens: A list of classification Tokens
        :param List transcripts: A list of transcript accessions
        :param Classification classification: A classification for a list of
            tokens
        :param List results: Stores validation result objects
        :param List gene_tokens: List of GeneMatchTokens for a classification
        :param Dict mane_data_found: MANE Transcript information found
        :param bool is_identifier: `True` if identifier is given for exact
            location. `False` otherwise.
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Optional[Endpoint] endpoint_name: Then name of the endpoint being used
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        """
        raise NotImplementedError

    def get_reference_sequence(
            self, ac: str, start: int, end: int, errors: List,
            cds_start: int = None) -> Optional[str]:
        """Get deleted reference sequence.

        :param str ac: Accession
        :param int start: Start position
        :param int end: End position
        :param List errors: List of errors
        :param int cds_start: Coding start site
        :return: Reference sequence of nucleotides
        """
        if cds_start:
            start += cds_start
            if end is not None:
                end += cds_start

        if start and not end:
            ref_sequence, _ = self.seqrepo_access.get_reference_sequence(ac, start)
        elif start is not None and end is not None:
            ref_sequence, _ = self.seqrepo_access.get_reference_sequence(ac, start,
                                                                         end + 1)
        else:
            ref_sequence = None

        if not ref_sequence:
            errors.append("Unable to get reference sequence.")
        return ref_sequence

    def check_reference_sequence(self, t: str, s: Token, errors: List,
                                 cds_start: int = None) -> bool:
        """Check that reference sequence matches deleted sequence.

        :param str t: Accession
        :param Token s: Classification token
        :param List errors: List of errors
        :param int cds_start: Coding start site
        :return: `True` if ref_sequences matches deleted_sequence.
            `False` otherwise.
        """
        ref_sequence = self.get_reference_sequence(
            t, s.start_pos_del, s.end_pos_del, errors, cds_start=cds_start)

        if errors:
            return False
        else:
            if ref_sequence and s.deleted_sequence:
                if ref_sequence != s.deleted_sequence:
                    errors.append(f"Expected deleted sequence {ref_sequence} "
                                  f"but got {s.deleted_sequence}")
                    return False
        return True

    async def add_normalized_genomic_dup_del(
            self, s: Token, t: str, start: int, end: int, gene: str,
            so_id: str, errors: List, hgvs_dup_del_mode: HGVSDupDelModeEnum,
            mane_data_found: Dict, baseline_copies: Optional[int] = None,
            relative_copy_class: Optional[RelativeCopyClass] = None) -> None:
        """Add normalized genomic dup or del to mane data

        :param Token s: Classification token
        :param str t: Accession
        :param int start: Start position
        :param int end: ENd position
        :param str gene: Gene
        :param str so_id: Sequence ontology id
        :param List errors: List of errors
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Dict mane_data_found: MANE Transcript information found
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        """
        mane = await self.mane_transcript.get_mane_transcript(
            t, start, s.coordinate_type, end_pos=end, gene=gene,
            try_longest_compatible=True)

        if mane:
            s.coordinate_type = "c"
            s.molecule_context = "transcript"
            s.so_id = so_id

            allele = self.vrs.to_vrs_allele(
                mane["refseq"], mane["pos"][0] + 1, mane["pos"][1] + 1,
                s.coordinate_type, s.alt_type, errors,
                cds_start=mane["coding_start_site"])

            mane_variation = self.hgvs_dup_del_mode.interpret_variation(
                t, s.alt_type, allele, errors, hgvs_dup_del_mode,
                baseline_copies=baseline_copies,
                relative_copy_class=relative_copy_class)

            if mane_variation:
                self._add_dict_to_mane_data(mane["refseq"], s, mane_variation,
                                            mane_data_found, mane["status"])

    async def validate_gene_or_accession_pos(
        self, t: str, pos_list: List, errors: List, gene: Optional[str] = None
    ) -> None:
        """Validate positions on gene or accession.
        If not valid, add to list of errors

        :param str t: Accession
        :param List pos_list: List of positions to validate
        :param List errors: List of errors
        :param Optional[str] gene: Gene
        """
        if gene:
            len_pos_list = len(pos_list)
            pos1, pos2 = pos_list[0], pos_list[1]
            if len_pos_list == 2:
                pos3, pos4 = None, None
            elif len_pos_list == 4:
                pos3, pos4 = pos_list[2], pos_list[3]
            else:
                errors.append(f"Unexpected amount of positions: {len_pos_list}")
                return
            await self._validate_gene_pos(gene, t, pos1, pos2, errors, pos3=pos3,
                                          pos4=pos4)
        else:
            for pos in pos_list:
                self._check_index(t, pos, errors)

    async def get_grch38_pos_ac(
            self, t: str, pos1: int, pos2: int,
            pos3: Optional[int] = None, pos4: Optional[int] = None
    ) -> Tuple[str, int, int, Optional[int], Optional[int], Optional[Dict]]:
        """Get GRCh38 start/end and accession

        :param str t: Accession
        :param int pos1: Position 1 (assumes 1-based)
        :param int pos2: Position 2 (assumes 1-based)
        :param int pos3: Position 3 (assumes 1-based)
        :param int pos4: Position 4 (assumes 1-based)
        :return: GRCh38 data (accession and positions)
        """
        grch38_start = await self.mane_transcript.g_to_grch38(t, pos1, pos2)
        if grch38_start:
            pos1, pos2 = grch38_start["pos"]
            if pos3 is not None and pos4 is not None:
                grch38_end = await self.mane_transcript.g_to_grch38(t, pos3, pos4)
                if grch38_end:
                    pos3, pos4 = grch38_end["pos"]
            t = grch38_start["ac"]
        return t, pos1, pos2, pos3, pos4, grch38_start

    def add_grch38_to_mane_data(
            self, t: str, s: Token, errors: List,
            grch38: Dict, mane_data_found: Dict,
            hgvs_dup_del_mode: HGVSDupDelModeEnum,
            ival: Optional[Tuple] = None,
            use_vrs_allele_range: bool = True,
            baseline_copies: Optional[int] = None,
            relative_copy_class: Optional[RelativeCopyClass] = None) -> None:
        """Add grch38 variation to mane data

        :param str t: Accession
        :param Token s: Classification token
        :param List errors: List of errors
        :param Dict grch38: GRCh38 data
        :param Dict mane_data_found: MANE data found for initial query
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Optionals[Tuple] ival: Interval
        :param bool use_vrs_allele_range: `True` if allele should be computed
            using `to_vrs_allele_ranges` method. `False` if allele should be
            computed using `to_vrs_allele` method.
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        """
        if errors:
            return

        if grch38:
            t = grch38["ac"]

        if use_vrs_allele_range:
            allele = self.vrs.to_vrs_allele_ranges(
                t, s.coordinate_type, s.alt_type, errors, ival)
        else:
            allele = self.vrs.to_vrs_allele(
                t, grch38["pos"][0], grch38["pos"][1], s.coordinate_type,
                s.alt_type, errors)

        grch38_variation = self.hgvs_dup_del_mode.interpret_variation(
            t, s.alt_type, allele, errors, hgvs_dup_del_mode,
            baseline_copies=baseline_copies, relative_copy_class=relative_copy_class)

        if grch38_variation:
            self._add_dict_to_mane_data(
                grch38["ac"], s, grch38_variation, mane_data_found, "GRCh38")



"""The module for Genomic Deletion Validation."""
import logging
from typing import List, Optional, Dict

from ga4gh.vrsatile.pydantic.vrs_models import RelativeCopyClass

from variation.schemas.app_schemas import Endpoint
from variation.validators.duplication_deletion_base import\
    DuplicationDeletionBase
from variation.schemas.classification_response_schema import \
    ClassificationType, Classification
from variation.schemas.token_response_schema import Token, SequenceOntology
from variation.schemas.token_response_schema import GeneMatchToken
from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum


logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)


class GenomicDeletion(DuplicationDeletionBase):
    """The Genomic Deletion Validator class."""

    async def get_transcripts(self, gene_tokens: List, classification: Classification,
                              errors: List) -> Optional[List[str]]:
        """Get transcript accessions for a given classification.

        :param List gene_tokens: A list of gene tokens
        :param Classification classification: A classification for a list of
            tokens
        :param List errors: List of errors
        :return: List of transcript accessions
        """
        transcripts = await self.get_genomic_transcripts(classification, errors)
        return transcripts

    async def get_valid_invalid_results(
        self, classification_tokens: List, transcripts: List,
        classification: Classification, results: List, gene_tokens: List,
        mane_data_found: Dict, is_identifier: bool,
        hgvs_dup_del_mode: HGVSDupDelModeEnum,
        endpoint_name: Optional[Endpoint] = None,
        baseline_copies: Optional[int] = None,
        relative_copy_class: Optional[RelativeCopyClass] = None,
        do_liftover: bool = False
    ) -> None:
        """Add validation result objects to a list of results.

        :param List classification_tokens: A list of classification Tokens
        :param List transcripts: A list of transcript accessions
        :param Classification classification: A classification for a list of
            tokens
        :param List results: Stores validation result objects
        :param List gene_tokens: List of GeneMatchTokens for a classification
        :param Dict mane_data_found: MANE Transcript information found
        :param bool is_identifier: `True` if identifier is given for exact
            location. `False` otherwise.
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Optional[Endpoint] endpoint_name: Then name of the endpoint being used
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        """
        valid_alleles = list()
        for s in classification_tokens:
            for t in transcripts:
                errors = list()
                t = self.get_accession(t, classification)
                start, end = s.start_pos_del, None
                if s.end_pos_del is None:
                    end = start
                else:
                    end = s.end_pos_del

                # Validate pos
                if gene_tokens:
                    await self.validate_gene_or_accession_pos(
                        t, [start, end], errors, gene=gene_tokens[0].token)
                else:
                    self.check_reference_sequence(t, s, errors)

                if not errors:
                    allele = self.vrs.to_vrs_allele(
                        t, s.start_pos_del, s.end_pos_del,
                        s.coordinate_type, s.alt_type, errors)

                    variation = self.hgvs_dup_del_mode.interpret_variation(
                        t, s.alt_type, allele, errors, hgvs_dup_del_mode,
                        pos=(start, end), relative_copy_class=relative_copy_class,
                        baseline_copies=baseline_copies
                    )
                else:
                    variation = None

                if not errors and (endpoint_name == Endpoint.NORMALIZE or do_liftover):
                    await self._get_normalize_variation(
                        gene_tokens, s, t, errors, hgvs_dup_del_mode,
                        mane_data_found, start, end,
                        relative_copy_class=relative_copy_class,
                        baseline_copies=baseline_copies)

                self.add_validation_result(
                    variation, valid_alleles, results,
                    classification, s, t, gene_tokens, errors
                )

                if is_identifier:
                    break

        if endpoint_name == Endpoint.NORMALIZE or do_liftover:
            self.add_mane_to_validation_results(
                mane_data_found, valid_alleles, results,
                classification, gene_tokens
            )

    async def _get_normalize_variation(
            self, gene_tokens: List, s: Token, t: str, errors: List,
            hgvs_dup_del_mode: HGVSDupDelModeEnum, mane_data_found: Dict,
            start: int, end: int,
            relative_copy_class: Optional[RelativeCopyClass] = None,
            baseline_copies: Optional[int] = None) -> None:
        """Get variation that will be returned in normalize endpoint.

        :param List gene_tokens: List of gene tokens
        :param Token s: Classification token
        :param str t: Accession
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Mode to use for
            interpreting HGVS duplications and deletions
        :param Dict mane_data_found: MANE Transcript data found for given query
        :param int start: Start pos change
        :param int end: End pos change
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param Optional[int] baseline_copies: Baseline copies number
        """
        if gene_tokens:
            await self.add_normalized_genomic_dup_del(
                s, t, s.start_pos_del, s.end_pos_del, gene_tokens[0].token,
                SequenceOntology.DELETION, errors, hgvs_dup_del_mode,
                mane_data_found, relative_copy_class=relative_copy_class,
                baseline_copies=baseline_copies)
        else:
            # No gene provided, then use GRCh38 assesmbly
            if not self._is_grch38_assembly(t):
                grch38 = await self.mane_transcript.g_to_grch38(t, start, end)
            else:
                grch38 = dict(ac=t, pos=(start, end))

            if grch38:
                await self.validate_gene_or_accession_pos(
                    grch38["ac"], [grch38["pos"][0], grch38["pos"][1]], errors)
                self.add_grch38_to_mane_data(
                    t, s, errors, grch38, mane_data_found, hgvs_dup_del_mode,
                    use_vrs_allele_range=False, relative_copy_class=relative_copy_class,
                    baseline_copies=baseline_copies)

    def get_gene_tokens(
            self, classification: Classification) -> List[GeneMatchToken]:
        """Return gene tokens for a classification.

        :param Classification classification: The classification for tokens
        :return: A list of Gene Match Tokens in the classification
        """
        return self.get_gene_symbol_tokens(classification)

    def variation_name(self) -> str:
        """Return the variation name."""
        return "genomic deletion"

    def is_token_instance(self, t: Token) -> bool:
        """Check that token is Genomic Deletion.

        :param Token t: Classification token
        """
        return t.token_type == "GenomicDeletion"

    def validates_classification_type(
            self, classification_type: ClassificationType) -> bool:
        """Return whether or not the classification type is
        Genomic DelIns.

        :param ClassificationType classification_type: Classification type
        :return: `True` if classification type matches, `False` otherwise
        """
        return classification_type == ClassificationType.GENOMIC_DELETION



"""A module for testing Coding DNA Insertion Tokenization."""
import unittest

from variation.tokenizers.caches import AminoAcidCache, NucleotideCache
from variation.tokenizers import CodingDNAInsertion
from .tokenizer_base import TokenizerBase


class TestCodingDNAInsertionTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Coding DNA Insertion Tokenization."""

    def tokenizer_instance(self):
        """Return Coding DNA Insertion instance."""
        return CodingDNAInsertion(AminoAcidCache(), NucleotideCache())

    def token_type(self):
        """Return Coding DNA insertion token type."""
        return "CodingDNAInsertion"

    def fixture_name(self):
        """Return the fixture name for Coding DNA Insertion."""
        return "coding_dna_insertion"



"""Module for testing gene pair tokenization."""
import unittest

from variation.tokenizers import GenePair
from .tokenizer_base import TokenizerBase


class TestGenePairTokenizer(TokenizerBase, unittest.TestCase):
    """Gene Pair Tokenizer Test class."""

    def tokenizer_instance(self):
        """Return Gene Pair Tokenizer instance."""
        return GenePair()

    def token_type(self):
        """Return Gene Pair token type."""
        return "GenePair"

    def fixture_name(self):
        """Return Gene Pair fixture name."""
        return "gene_pair"



"""Module for testing Coding DNA Deletion Validator."""
import unittest

from variation.validators import CodingDNADeletion
from variation.classifiers import CodingDNADeletionClassifier
from .validator_base import ValidatorBase


class TestCodingDNADeletionValidator(ValidatorBase, unittest.TestCase):
    """A class to test the CodingDNADeletion Validator."""

    def validator_instance(self):
        """Return coding dna deletion instance."""
        return CodingDNADeletion(*self.params)

    def classifier_instance(self):
        """Return the coding dna deletion classifier instance."""
        return CodingDNADeletionClassifier()

    def fixture_name(self):
        """Return the fixture name for coding dna deletion."""
        return "coding_dna_deletion"



"""Module for testing Genomic Deletion Validator."""
import unittest

from variation.validators import GenomicDeletion
from variation.classifiers import GenomicDeletionClassifier
from .validator_base import ValidatorBase


class TestGenomicDeletionValidator(ValidatorBase, unittest.TestCase):
    """A class to test the GenomicDeletion Validator."""

    def validator_instance(self):
        """Return genomic deletion instance."""
        return GenomicDeletion(*self.params)

    def classifier_instance(self):
        """Return the genomic deletion classifier instance."""
        return GenomicDeletionClassifier()

    def fixture_name(self):
        """Return the fixture name for genomic deletion."""
        return "genomic_deletion"



"""A module for testing Genomic Deletion Ranges tokenization."""
import unittest

from variation.tokenizers import GenomicDeletionRange
from .tokenizer_base import TokenizerBase


class TestGenomicDeletionRangeTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Genomic Deletion Range Tokenization."""

    def tokenizer_instance(self):
        """Return Genomic Deletion Range instance."""
        return GenomicDeletionRange()

    def token_type(self):
        """Return genomic deletion range token type."""
        return "GenomicDeletionRange"

    def fixture_name(self):
        """Return the fixture name for genomic deletion range."""
        return "genomic_deletion_range"



"""Module for testing Genomic Duplication Validator."""
import unittest

from variation.validators import GenomicDuplication
from variation.classifiers import GenomicDuplicationClassifier
from .validator_base import ValidatorBase


class TestGenomicDuplicationValidator(ValidatorBase, unittest.TestCase):
    """A class to test the Genomic Duplication Validator."""

    def validator_instance(self):
        """Return genomic duplication instance."""
        return GenomicDuplication(*self.params)

    def classifier_instance(self):
        """Return the genomic duplication classifier instance."""
        return GenomicDuplicationClassifier()

    def fixture_name(self):
        """Return the fixture name for genomic duplication."""
        return "genomic_duplication"



"""Module for Genomic Silent Mutation Translation."""
from variation.translators.translator import Translator
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import GenomicSilentMutationToken, Token


class GenomicSilentMutation(Translator):
    """The Genomic Silent Mutation Translator class."""

    def can_translate(self, type: ClassificationType) -> bool:
        """Return if classification type is Genomic Silent Mutation."""
        return type == ClassificationType.GENOMIC_SILENT_MUTATION

    def is_token_instance(self, token: Token) -> bool:
        """Return if the token is an Genomic Silent Mutation token
        instance.
        """
        return isinstance(token, GenomicSilentMutationToken)



"""Validator package level import."""
from .validate import Validate
from .validator import Validator
from .polypeptide_sequence_variation_base import PolypeptideSequenceVariationBase
from .protein_substitution import ProteinSubstitution
from .polypeptide_truncation import PolypeptideTruncation
from .silent_mutation import SilentMutation
from .single_nucleotide_variation_base import SingleNucleotideVariationBase
from .coding_dna_substitution import CodingDNASubstitution
from .genomic_substitution import GenomicSubstitution
from .coding_dna_silent_mutation import CodingDNASilentMutation
from .genomic_silent_mutation import GenomicSilentMutation
from .protein_delins import ProteinDelIns
from .coding_dna_delins import CodingDNADelIns
from .genomic_delins import GenomicDelIns
from .protein_deletion import ProteinDeletion
from .coding_dna_deletion import CodingDNADeletion
from .genomic_deletion import GenomicDeletion
from .genomic_base import GenomicBase
from .protein_insertion import ProteinInsertion
from .coding_dna_insertion import CodingDNAInsertion
from .genomic_insertion import GenomicInsertion
from .genomic_uncertain_deletion import GenomicUncertainDeletion
from .genomic_duplication import GenomicDuplication
from .genomic_deletion_range import GenomicDeletionRange



"""A module for Coding DNA DelIns Tokenization."""
from typing import Dict, Optional

from variation.schemas.token_response_schema import CodingDNADelInsToken
from variation.tokenizers.delins_base import DelInsBase


class CodingDNADelIns(DelInsBase):
    """Class for tokenizing DelIns at the coding dna reference sequence."""

    def return_token(self, params: Dict) -> Optional[CodingDNADelInsToken]:
        """Return coding DNA DelIns token."""
        if self.parts["coordinate_type"] == "c":
            return CodingDNADelInsToken(**params)



"""Module for Token Schema."""
from typing import List, Union, Dict, Any, Type, Optional, Literal
from enum import IntEnum, Enum

from pydantic import BaseModel


class TokenMatchType(IntEnum):
    """Gene Token match types."""

    ID = 1
    SYMBOL = 2
    PREVIOUS = 3
    ALIAS = 4
    UNSPECIFIED = 5


class TokenType(str, Enum):
    """Define token types."""

    # TODO: Add other token types (issue 162)
    GENOMIC_DUPLICATION = "GenomicDuplication"
    GENOMIC_DUPLICATION_RANGE = "GenomicDuplicationRange"


class Nomenclature(str, Enum):
    """Define nomenclatures that are supported"""

    FREE_TEXT = "free_text"
    HGVS = "hgvs"
    GNOMAD_VCF = "gnomad_vcf"


class SequenceOntology(str, Enum):
    """Define SequenceOntology codes"""

    POLYPEPTIDE_TRUNCATION = "SO:0001617"
    PROTEIN_SUBSTITUTION = "SO:0001606"
    SILENT_MUTATION = "SO:0001017"
    SNV = "SO:0001483"
    NO_SEQUENCE_ALTERATION = "SO:0002073"
    DELINS = "SO:1000032"
    PROTEIN_INSERTION = "SO:0001605"
    INSERTION = "SO:0000667"
    PROTEIN_DELETION = "SO:0001604"
    DELETION = "SO:0000159"
    COPY_NUMBER_LOSS = "SO:0001743"
    DUPLICATION = "SO:1000035"
    COPY_NUMBER_GAIN = "SO:0001742"


class Token(BaseModel):
    """A string from a given query."""

    token: str
    token_type: str
    match_type: TokenMatchType
    input_string: str
    object_type = "Token"
    nomenclature: Optional[Nomenclature]

    class Config:
        """Configure model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["Token"]) -> None:
            """Configure OpenAPI schema."""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "token": "",
                "token_type": "Unknown",
                "match_type": 5,
                "input_string": "foo",
                "object_type": "Token",
                "nomenclature": None
            }


class GeneMatchToken(Token):
    """Define model for gene symbol token."""

    matched_value: str
    token_type = "GeneSymbol"

    class Config:
        """Configure model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["GeneMatchToken"]) -> None:
            """Configure OpenAPI schema."""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "token": "BRAF",
                "token_type": "GeneSymbol",
                "match_type": 2,
                "input_string": "BRAF",
                "object_type": "Token",
                "matched_value": "BRAF"
            }


class GenePairMatchToken(Token):
    """Define model for gene pair token."""

    left_gene_token: GeneMatchToken
    right_gene_token: GeneMatchToken
    token_type = "GenePair"

    class Config:
        """Configure model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["GenePairMatchToken"]) -> None:
            """Configure OpenAPI schema."""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "token": "BRAF-ABL1",
                "token_type": "GenePair",
                "match_type": 5,
                "input_string": "braf-abl1",
                "object_type": "Token",
                "left_gene_token": {
                    "token": "BRAF",
                    "token_type": "GeneSymbol",
                    "match_type": 2,
                    "input_string": "BRAF",
                    "object_type": "Token",
                    "matched_value": "BRAF"
                },
                "right_gene_token": {
                    "token": "ABL1",
                    "token_type": "GeneSymbol",
                    "match_type": 2,
                    "input_string": "ABL1",
                    "object_type": "Token",
                    "matched_value": "ABL1"
                }
            }


class CoordinateType(str, Enum):
    """Define constraints for coordinate types."""

    CODING_DNA = "c"
    LINEAR_GENOMIC = "g"
    PROTEIN = "p"


class PolypeptideSequenceVariation(Token):
    """Polypeptide Sequence Variation Token Class."""

    ref_protein: str
    alt_protein: str
    position: int
    token_type: str
    coordinate_type = CoordinateType.PROTEIN
    so_id: SequenceOntology
    molecule_context = "protein"
    alt_type: str


class PolypeptideTruncationToken(PolypeptideSequenceVariation):
    """A sequence variation of the CD that causes a truncation of the
    resulting polypeptide. (nonsense)
    """

    alt_protein = "*"
    token_type = "PolypeptideTruncation"
    so_id = SequenceOntology.POLYPEPTIDE_TRUNCATION
    alt_type = "nonsense"

    class Config:
        """Configure model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["PolypeptideTruncationToken"]) -> None:
            """Configure OpenAPI schema."""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "token": "Tyr365Ter",
                "token_type": "PolypeptideTruncation",
                "match_type": 5,
                "input_string": "Tyr365Ter",
                "object_type": "Token",
                "ref_protein": "Tyr",
                "alt_protein": "Ter",
                "position": 365
            }


class ProteinSubstitutionToken(PolypeptideSequenceVariation):
    """A sequence variation of a codon resulting in the substitution of one
    amino acid for another in the resulting polypeptide. (missense)
    """

    token_type = "ProteinSubstitution"
    so_id = SequenceOntology.PROTEIN_SUBSTITUTION
    alt_type = "substitution"

    class Config:
        """Configure model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["ProteinSubstitutionToken"]) -> None:
            """Configure OpenAPI schema."""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "token": "V600E",
                "token_type": "ProteinSubstitution",
                "match_type": 5,
                "input_string": "V600E",
                "object_type": "Token",
                "ref_protein": "V",
                "alt_protein": "E",
                "position": 600
            }


class SilentMutationToken(PolypeptideSequenceVariation):
    """A sequence variation that does not affect protein functions."""

    alt_protein = "="
    token_type = "SilentMutation"
    so_id = SequenceOntology.SILENT_MUTATION
    alt_type = "silent_mutation"

    class Config:
        """Configure model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["SilentMutationToken"]) -> None:
            """Configure OpenAPI schema."""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "token": "p.Cys188=",
                "token_type": "SilentMutation",
                "match_type": 5,
                "input_string": "p.Cys188=",
                "object_type": "Token",
                "ref_protein": "Cys",
                "alt_protein": "=",
                "position": 188
            }


class TokenResponseSchema(BaseModel):
    """Define model for token response."""

    search_term: str
    tokens: List[Union[GeneMatchToken, GenePairMatchToken,
                       ProteinSubstitutionToken, PolypeptideTruncationToken,
                       SilentMutationToken, Token]]

    class Config:
        """Configure model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["TokenResponseSchema"]) -> None:
            """Configure OpenAPI schema."""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "search_term": "BRAF V600E",
                "tokens": [
                    {
                        "token": "BRAF",
                        "token_type": "GeneSymbol",
                        "match_type": 2,
                        "input_string": "BRAF",
                        "object_type": "Token",
                        "matched_value": "BRAF"
                    },
                    {
                        "token": "V600E",
                        "token_type": "ProteinSubstitution",
                        "match_type": 5,
                        "input_string": "V600E",
                        "object_type": "Token",
                        "ref_protein": "V",
                        "alt_protein": "E",
                        "position": 600
                    }
                ]
            }


class SingleNucleotideVariation(Token):
    """Single nucleotide positions in genomic DNA at which different
    sequence alternatives exist.
    """

    position: int
    ref_nucleotide: Optional[str]
    new_nucleotide: str
    token_type: str
    coordinate_type: CoordinateType
    so_id: SequenceOntology
    molecule_context: str
    alt_type: str


class CodingDNASubstitutionToken(SingleNucleotideVariation):
    """SNV substitution at the coding DNA reference sequence."""

    coordinate_type = CoordinateType.CODING_DNA
    token_type = "CodingDNASubstitution"
    so_id = SequenceOntology.SNV
    molecule_context = "transcript"
    alt_type = "substitution"


class CodingDNASilentMutationToken(SingleNucleotideVariation):
    """SNV no change at the coding DNA reference sequence."""

    coordinate_type = CoordinateType.CODING_DNA
    new_nucleotide = "="
    token_type = "CodingDNASilentMutation"
    so_id = SequenceOntology.NO_SEQUENCE_ALTERATION
    molecule_context = "transcript"
    alt_type = "silent_mutation"


class GenomicSubstitutionToken(SingleNucleotideVariation):
    """SNV substitution at the linear genomic reference sequence."""

    coordinate_type = CoordinateType.LINEAR_GENOMIC
    token_type = "GenomicSubstitution"
    so_id = SequenceOntology.SNV
    molecule_context = "genomic"
    alt_type = "substitution"


class GenomicSilentMutationToken(SingleNucleotideVariation):
    """SNV no change at the linear genomic reference sequence."""

    coordinate_type = CoordinateType.LINEAR_GENOMIC
    new_nucleotide = "="
    token_type = "GenomicSilentMutation"
    so_id = SequenceOntology.NO_SEQUENCE_ALTERATION
    molecule_context = "genomic"
    alt_type = "silent_mutation"


class DelIns(Token):
    """A sequence alteration which included an insertion and a deletion,
    affecting 2 or more bases.
    """

    start_pos_del: str
    end_pos_del: Optional[str]
    inserted_sequence1: str
    inserted_sequence2: Optional[str]
    token_type: str
    coordinate_type: CoordinateType
    so_id = SequenceOntology.DELINS
    molecule_context: str
    alt_type = "delins"


class ProteinDelInsToken(Token):
    """DelIns at the protein reference sequence."""

    start_aa_del: str
    start_pos_del: int
    end_aa_del: Optional[str]
    end_pos_del: Optional[int]
    inserted_sequence: str
    coordinate_type = CoordinateType.PROTEIN
    so_id = SequenceOntology.DELINS
    molecule_context = "protein"
    token_type = "ProteinDelIns"
    alt_type = "delins"


class CodingDNADelInsToken(DelIns):
    """DelIns at the coding DNA reference sequence."""

    coordinate_type = CoordinateType.CODING_DNA
    token_type = "CodingDNADelIns"
    molecule_context = "transcript"


class GenomicDelInsToken(DelIns):
    """DelIns at the linear genomic reference sequence."""

    coordinate_type = CoordinateType.LINEAR_GENOMIC
    token_type = "GenomicDelIns"
    molecule_context = "genomic"


class LocusReferenceGenomicToken(Token):
    """Contain stable reference sequences that are used for reporting
    sequence variations with clinical implications.
    """

    id: int
    t: Optional[int]
    p: Optional[int]
    token_type = "LocusReferenceGenomic"


class Insertion(Token):
    """a sequence change between the translation initiation (start) and
    termination (stop) codon where, compared to the reference sequence,
    one or more amino acids are inserted, which is not a frame shift and
    where the insertion is not a copy of a sequence immediately N-terminal
    (5') - varnomen.hgvs.org
    """

    start_pos_flank: int
    end_pos_flank: int
    inserted_sequence: str
    coordinate_type: CoordinateType
    token_type: str
    so_id: SequenceOntology
    molecule_context: str
    alt_type = "insertion"


class ProteinInsertionToken(Insertion):
    """Protein Insertion."""

    start_aa_flank: str
    end_aa_flank: str
    coordinate_type = CoordinateType.PROTEIN
    token_type = "ProteinInsertion"
    so_id = SequenceOntology.PROTEIN_INSERTION
    molecule_context = "protein"


class CodingDNAInsertionToken(Insertion):
    """Coding DNA Insertion."""

    coordinate_type = CoordinateType.CODING_DNA
    inserted_sequence2: Optional[str]
    token_type = "CodingDNAInsertion"
    so_id = SequenceOntology.INSERTION
    molecule_context = "transcript"


class GenomicInsertionToken(Insertion):
    """Genomic Insertion."""

    coordinate_type = CoordinateType.LINEAR_GENOMIC
    inserted_sequence2: Optional[str]
    token_type = "GenomicInsertion"
    so_id = SequenceOntology.INSERTION
    molecule_context = "genomic"


class DeletionAltType(str, Enum):
    """Define alt types for deletions."""

    DELETION = "deletion"
    DELETION_RANGE = "deletion_range"
    UNCERTAIN_DELETION = "uncertain_deletion"


class Deletion(Token):
    """The point at which one or more contiguous nucleotides were excised.
    - Sequence Ontology
    """

    start_pos_del: int
    end_pos_del: Optional[int]
    coordinate_type: CoordinateType
    token_type: str
    so_id: SequenceOntology
    molecule_context: str
    alt_type: DeletionAltType.DELETION = DeletionAltType.DELETION


class ProteinDeletionToken(Deletion):
    """A sequence change between the translation initiation (start) and
    termination (stop) codon where, compared to a reference sequence, one or
    more amino acids are not present (deleted) - varnomen.hgvs.org
    """

    start_aa_del: str
    end_aa_del: Optional[str]
    coordinate_type = CoordinateType.PROTEIN
    token_type = "ProteinDeletion"
    so_id = SequenceOntology.PROTEIN_DELETION
    molecule_context = "protein"


class CodingDNADeletionToken(Deletion):
    """A sequence change where, compared to a reference sequence, one or
    more nucleotides are not present (deleted). - varnomen.hgvs.org
    """

    coordinate_type = CoordinateType.CODING_DNA
    deleted_sequence: Optional[str]
    token_type = "CodingDNADeletion"
    so_id = SequenceOntology.DELETION
    molecule_context = "transcript"


class GenomicDeletionToken(Deletion):
    """A sequence change where, compared to a reference sequence, one or
    more nucleotides are not present (deleted). - varnomen.hgvs.org
    """

    coordinate_type = CoordinateType.LINEAR_GENOMIC
    deleted_sequence: Optional[str]
    token_type = "GenomicDeletion"
    so_id = SequenceOntology.DELETION
    molecule_context = "genomic"


class DeletionRange(Token):
    """Deletions of the form (pos_pos)_(pos_pos)."""

    start_pos1_del: Union[int, str]
    start_pos2_del: int
    end_pos1_del: int
    end_pos2_del: Union[int, str]
    token_type: str
    so_id = SequenceOntology.COPY_NUMBER_LOSS
    molecule_context: str
    alt_type: Union[Literal[DeletionAltType.DELETION_RANGE], Literal[DeletionAltType.UNCERTAIN_DELETION]] = DeletionAltType.DELETION_RANGE  # noqa: E501


class GenomicDeletionRangeToken(DeletionRange):
    """Genomic deletion range token."""

    token_type = "GenomicDeletionRange"
    molecule_context = "genomic"
    coordinate_type = CoordinateType.LINEAR_GENOMIC


class UncertainDeletion(DeletionRange):
    """Uncertain Deletion."""

    start_pos1_del: Optional[Union[Literal["?"], int]]
    start_pos2_del: Optional[int]
    end_pos1_del: int
    end_pos2_del: Optional[Union[Literal["?"], int]]
    token_type: str
    molecule_context: str
    alt_type: Literal[DeletionAltType.UNCERTAIN_DELETION] = DeletionAltType.UNCERTAIN_DELETION  # noqa: E501


class GenomicUncertainDeletionToken(UncertainDeletion):
    """Genomic uncertain deletion."""

    token_type = "GenomicUncertainDeletion"
    molecule_context = "genomic"
    coordinate_type = CoordinateType.LINEAR_GENOMIC


class DuplicationAltType(str, Enum):
    """Define alt types for duplications."""

    DUPLICATION = "duplication"
    DUPLICATION_RANGE = "duplication_range"
    UNCERTAIN_DUPLICATION = "uncertain_duplication"


class Duplication(Token):
    """Duplications."""

    start_pos1_dup: Union[Literal["?"], int]
    start_pos2_dup: Optional[int]
    token_type: TokenType
    so_id = SequenceOntology.DUPLICATION
    molecule_context: str
    alt_type: DuplicationAltType


class GenomicDuplicationToken(Duplication):
    """Genomic duplication token schema."""

    token_type = TokenType.GENOMIC_DUPLICATION
    molecule_context = "genomic"
    coordinate_type = CoordinateType.LINEAR_GENOMIC
    alt_type: Literal[DuplicationAltType.DUPLICATION] = DuplicationAltType.DUPLICATION  # noqa: E501


class DuplicationRange(Duplication):
    """Duplications of the form (#_#)_(#_#)dup"""

    end_pos1_dup: int
    end_pos2_dup: Optional[Union[Literal["?"], int]]
    so_id = SequenceOntology.COPY_NUMBER_GAIN


class GenomicDuplicationRangeToken(DuplicationRange):
    """Genomic Duplication Range token schema"""

    token_type = TokenType.GENOMIC_DUPLICATION_RANGE
    molecule_context = "genomic"
    coordinate_type = CoordinateType.LINEAR_GENOMIC


class ChromosomeToken(Token):
    """Chromosome token"""

    chromosome: str
    token_type = "Chromosome"



"""Module for normalize endpoint response schema."""
from enum import Enum
from typing import List, Optional, Dict, Any, Type, Union
from datetime import datetime

from pydantic import BaseModel
from pydantic.types import StrictStr
from ga4gh.vrsatile.pydantic.vrsatile_models import VariationDescriptor, \
    CanonicalVariation, ComplexVariation


class HGVSDupDelMode(str, Enum):
    """Define options for HGVSDupDelMode.
    This mode determines how to interpret HGVS dup/del.
    """

    DEFAULT = "default"
    CNV = "cnv"
    REPEATED_SEQ_EXPR = "repeated_seq_expr"
    LITERAL_SEQ_EXPR = "literal_seq_expr"


class ServiceMeta(BaseModel):
    """Metadata regarding the variation-normalization service."""

    name = "variation-normalizer"
    version: StrictStr
    response_datetime: datetime
    url = "https://github.com/cancervariants/variation-normalization"

    class Config:
        """Configure schema example."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["ServiceMeta"]) -> None:
            """Configure OpenAPI schema"""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "name": "variation-normalizer",
                "version": "0.1.0",
                "response_datetime": "2021-04-05T16:44:15.367831",
                "url": "https://github.com/cancervariants/variation-normalization"  # noqa: E501
            }


class ServiceResponse(BaseModel):
    """Base response model for services"""

    warnings: Optional[List[StrictStr]]
    service_meta_: ServiceMeta


class NormalizeService(ServiceResponse):
    """A response to normalizing a variation to a single GA4GH Value Object Descriptor."""  # noqa: E501

    variation_query: StrictStr
    variation_descriptor: Optional[VariationDescriptor]

    class Config:
        """Configure model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["NormalizeService"]) -> None:
            """Configure OpenAPI schema."""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "variation_query": "BRAF V600E",
                "variation_descriptor": {
                    "id": "normalize.variation:BRAF%20V600E",
                    "label": "BRAF V600E",
                    "type": "VariationDescriptor",
                    "variation_id": "ga4gh:VA.8JkgnqIgYqufNl-OV_hpRG_aWF9UFQCE",  # noqa: E501
                    "variation": {
                        "_id": "ga4gh:VA.8JkgnqIgYqufNl-OV_hpRG_aWF9UFQCE",
                        "location": {
                            "_id": "ga4gh:VSL.AqrQ-EkAvTrXOFn70_8i3dXF5shBBZ5i",  # noqa: E501
                            "interval": {
                                "end": {"value": 640, "type": "Number"},
                                "start": {"value": 639, "type": "Number"},
                                "type": "SequenceInterval"
                            },
                            "sequence_id": "ga4gh:SQ.WaAJ_cXXn9YpMNfhcq9lnzIvaB9ALawo",  # noqa: E501
                            "type": "SequenceLocation"
                        },
                        "state": {
                            "sequence": "E",
                            "type": "LiteralSequenceExpression"
                        },
                        "type": "Allele"
                    },
                    "molecule_context": "protein",
                    "structural_type": "SO:0001606",
                    "vrs_ref_allele_seq": "V",
                    "gene_context": {
                        "id": "normalize.gene:BRAF",
                        "type": "GeneDescriptor",
                        "label": "BRAF",
                        "gene_id": "hgnc:1097",
                        "xrefs": [
                            "ncbigene:673",
                            "ensembl:ENSG00000157764"
                        ],
                        "alternate_labels": [
                            "B-Raf proto-oncogene, serine/threonine kinase",
                            "BRAF1"
                        ],
                        "extensions": [
                            {
                                "name": "symbol_status",
                                "value": "approved",
                                "type": "Extension"
                            },
                            {
                                "name": "associated_with",
                                "value": [
                                    "vega:OTTHUMG00000157457",
                                    "ucsc:uc003vwc.5",
                                    "ccds:CCDS5863",
                                    "ccds:CCDS87555",
                                    "uniprot:P15056",
                                    "pubmed:2284096",
                                    "pubmed:1565476",
                                    "cosmic:BRAF",
                                    "omim:164757",
                                    "orphanet:119066",
                                    "iuphar:1943",
                                    "ena.embl:M95712",
                                    "refseq:NM_004333"
                                ],
                                "type": "Extension"
                            },
                            {
                                "name": "chromosome_location",
                                "value": {
                                    "_id": "ga4gh:VCL.O6yCQ1cnThOrTfK9YUgMlTfM6HTqbrKw",  # noqa: E501
                                    "type": "ChromosomeLocation",
                                    "species_id": "taxonomy:9606",
                                    "chr": "7",
                                    "interval": {
                                        "end": "q34",
                                        "start": "q34",
                                        "type": "CytobandInterval"
                                    }
                                },
                                "type": "Extension"
                            }
                        ]
                    }
                },
                "service_meta_": {
                    "name": "variation-normalizer",
                    "version": "0.2.17",
                    "response_datetime": "2022-01-26T22:23:41.821673",
                    "url": "https://github.com/cancervariants/variation-normalization"  # noqa: E501
                }
            }


class TranslateIdentifierService(ServiceResponse):
    """A response to translating identifiers."""

    identifier_query: StrictStr
    aliases: List[StrictStr]

    class Config:
        """Configure model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["TranslateIdentifierService"]) -> None:
            """Configure OpenAPI schema."""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "identifier_query": "NP_004324.2",
                "warnings": None,
                "aliases": [
                    "Ensembl:ENSP00000288602.6",
                    "ensembl:ENSP00000288602.6",
                    "Ensembl:ENSP00000493543.1",
                    "ensembl:ENSP00000493543.1",
                    "MD5:74c9b69323bd112084c1b5b385e7e6c5",
                    "NCBI:NP_004324.2",
                    "refseq:NP_004324.2",
                    "SEGUID:sfzILpNpX8UFB/vgH9LOKLpl/+g",
                    "SHA1:b1fcc82e93695fc50507fbe01fd2ce28ba65ffe8",
                    "VMC:GS_cQvw4UsHHRRlogxbWCB8W-mKD4AraM9y",
                    "sha512t24u:cQvw4UsHHRRlogxbWCB8W-mKD4AraM9y",
                    "ga4gh:SQ.cQvw4UsHHRRlogxbWCB8W-mKD4AraM9y"
                ],
                "service_meta_": {
                    "name": "variation-normalizer",
                    "version": "0.2.14",
                    "response_datetime": "2021-11-18T14:10:53.909158",
                    "url": "https://github.com/cancervariants/variation-normalization"  # noqa: E501
                }
            }


class CanonicalSPDIToCategoricalVariationService(ServiceResponse):
    """A response model for the canonical spdi to categorical variation service"""

    canonical_spdi_query: str
    categorical_variation: Optional[Union[CanonicalVariation, ComplexVariation]]

    class Config:
        """Configure model."""

        @staticmethod
        def schema_extra(
                schema: Dict[str, Any],
                model: Type["CanonicalSPDIToCategoricalVariationService"]) -> None:
            """Configure OpenAPI schema."""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "canonical_spdi_query": "NC_000007.14:140753335:A:T",
                "warnings": [],
                "categorical_variation": {
                    "_id": "ga4gh:VCC.W0r_NF_ecKXjgvTwcMNkyVS1pB_CXMj9",
                    "type": "CanonicalVariation",
                    "complement": False,
                    "variation": {
                        "_id": "ga4gh:VA.fZiBjQEolbkL0AxjoTZf4SOkFy9J0ebU",
                        "type": "Allele",
                        "location": {
                            "_id": "ga4gh:VSL.zga82-TpYiNmBESCfvDvAz9DyvJF98I-",
                            "type": "SequenceLocation",
                            "sequence_id": "ga4gh:SQ.F-LrLMe1SRpfUZHkQmvkVKFEGaoDeHul",
                            "interval": {
                                "type": "SequenceInterval",
                                "start": {
                                    "type": "Number",
                                    "value": 140753335
                                },
                                "end": {
                                    "type": "Number",
                                    "value": 140753336
                                }
                            }
                        },
                        "state": {
                            "type": "LiteralSequenceExpression",
                            "sequence": "T"
                        }
                    }
                },
                "service_meta_": {
                    "version": "0.2.20",
                    "response_datetime": "2022-02-20T17:16:19.415675",
                    "name": "variation-normalizer",
                    "url": "https://github.com/cancervariants/variation-normalization"
                }
            }



"""Module for testing Silent Mutation Translator."""
import unittest

from variation.classifiers import SilentMutationClassifier
from variation.translators import SilentMutation
from variation.validators import SilentMutation as SM_V
from .translator_base import TranslatorBase


class TestSilentMutationTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the silent mutation Translator."""

    def classifier_instance(self):
        """Return silent mutation instance."""
        return SilentMutationClassifier()

    def validator_instance(self):
        """Return silent mutation instance."""
        return SM_V(*self.aa_params)

    def translator_instance(self):
        """Return silent mutation instance."""
        return SilentMutation()

    def fixture_name(self):
        """Return the fixture name for silent mutation."""
        return "silent_mutation"
